local
    open Normsyn Util
    structure A = Absyn
    structure G = GrailAbsyn
    val () = Absyn.required
    val () = GrailAbsyn.required

    (* Associate name of class with names of superclass/ interfaces *)
    val superTable = ref (Binarymap.mkDict String.compare)
    val interfaceTable = ref (Binarymap.mkDict String.compare)
    fun saveSuper c (s,_) = superTable := Binarymap.insert(!superTable, c, s)
    fun saveInterfaces c l = interfaceTable :=
	      Binarymap.insert(!interfaceTable,c, map nameOf l)

    fun id x = x
in

fun printInterfaces () = Binarymap.app (fn (k,l) => println (k ^" --> ["
								 ^ listToString id ", " l
								 ^ "]")) (!interfaceTable)

fun cpError s = Util.ierror ("[Classpath.sml]: " ^ s)

datatype classholder = DIR of string | ZIP of ZipReader.zipfile

local
    val cp: string list option ref  = ref NONE

    val classHolders = ref []
    val loaded = ref false

    fun setClassPath' s = (* s is a colon-separated list of directory names *)
    let
	val separator =
	    if member #";" (String.explode s)
	       orelse Path.validVolume {isAbs=true, vol="A:"}
	    then #";"      (* Windows *)
	    else #":"      (* Unix *)

	val files = String.tokens (fn c => c = separator) s
        (* Note that we haven't trimmed trailing slashes; we always
           manipulate paths and filenames using the functions in Path,
           and these make sure that slashes (or other separators)
           are in the correct places *)
    in
	cp := SOME files
    end

    fun initClasspath () =
	case (!cp) of
	    NONE => (* not set from command-line, so look elsewhere *)
	    let
		val pwd = getOpt (Process.getEnv "PWD", ".")
		val default =
		    case Process.getEnv "CAMELOT_CLASSPATH"
		     of NONE => (* try ${HOME}/classes *)
			let in
			    case Process.getEnv "HOME"
			     of NONE => pwd
			      | SOME home =>
				let val f = Path.joinDirFile
						{dir=home, file="classes"}
				in if Nonstdio.file_exists f then f ^ ":" ^ pwd else pwd end
			end
		      | SOME s => s
	    in
		setClassPath' default
	    end
	  | SOME p => ()
in

fun classpath () = valOf (!cp)
fun setClassPath s = setClassPath' s


(* The next function can be quite time consuming,  so we don't want to call
   it unless we're definitely using OO stuff;  this is why the stuff with
   'loaded' is there. *)

fun getClassHolders () =
    if (!loaded) then !classHolders
    else
    let
	val () = loaded := true
	val () = initClasspath ()
	fun doFile f =
	    case (#ext (Path.splitBaseExt f)) of
		SOME "jar" => ZIP (ZipReader.open_in f)
	      |	SOME "zip" => ZIP (ZipReader.open_in f)
	      | _ => DIR f
    in
	classHolders := map doFile (valOf (!cp));
	!classHolders
    end

fun closeJars () = app (fn ZIP f => ZipReader.close_in f | _ => ()) (!classHolders)
end  (* local *)


local
    val knownClasses = ref (Binaryset.empty String.compare)
    val internalClasses = ref (Binaryset.empty String.compare)  (* Classes defined in current file *)
    val interfaceTable = ref (Splayset.empty String.compare)
    (* Table of names of "classes" which are really interfaces *)
in

fun markKnown c = knownClasses := Binaryset.add(!knownClasses,c)
fun isKnownClass c = Binaryset.member(!knownClasses,c)

fun markInternal (CLASSdec((cname,_),_,_,_)) =
    internalClasses := Binaryset.add(!internalClasses,cname)
fun isInternal (c,_) = Binaryset.member(!internalClasses,c)

fun recordInterface c = (interfaceTable := Splayset.add(!interfaceTable,c))
fun isInterface c = Splayset.member(!interfaceTable,c)  (* Needed in Phi.sml *)

end

fun gtyToTy ty = case ty of
    G.INTty => A.INTty
  | G.BYTEty => A.BYTEty
  | G.SHORTty => A.SHORTty
  | G.LONGty => A.LONGty
  | G.CHARty => A.CHARty
  | G.FLOATty => A.FLOATty
  | G.DOUBLEty => A.DOUBLEty
  | G.BOOLEANty => A.BOOLty
  | G.REFty "java.lang.String" => STRINGty
  | G.REFty name => OBJECTty name
  | G.ARRAYty ty => A.ARRAYty (gtyToTy ty)


fun cnameToFilename cname = String.translate (fn #"." => "/" | x => str x) cname
fun filenameToClass fname = String.translate (fn #"/" => "." | x => str x) fname

fun hasFlag flag flags = List.exists (fn x => x = flag) flags
fun f_public flags = hasFlag Classdecl.F_ACCpublic flags
fun f_static flags = if hasFlag Classdecl.F_ACCstatic flags then STATIC else INSTANCE
fun m_public flags = hasFlag Classdecl.M_ACCpublic flags
fun m_static flags = if hasFlag Classdecl.M_ACCstatic flags then STATIC else INSTANCE

open Jvmtype

fun qualName class =
    let
        fun f (a::b::t) = f ((a ^"."^b)::t)
          | f [x] = x
          | f [] = ""
        val toks = (Jvmtype.packages class)@[Jvmtype.className class]
    in f toks end

fun jtyToTy t =   (* Used to be in Compile; maybe somewhere better ? *)
    case t of
	Tint => G.INTty
      | Tboolean => G.BOOLEANty
      | Tfloat => G.FLOATty
      | Tclass c => G.REFty (qualName c)
      | Tarray ty => G.ARRAYty (jtyToTy ty)
      | Tchar => G.CHARty
      | Tbyte => G.BYTEty
      | Tshort => G.SHORTty
      | Tlong => G.LONGty
      | Tdouble => G.DOUBLEty

fun jtyToRTy (SOME t) = SOME (jtyToTy t)
  | jtyToRTy NONE = NONE


fun mdeclToValdec {flags, name, msig = (tys, rty), attrs} =
    let
         val (tysG, rtyG) = (map jtyToTy tys, jtyToRTy rty)
         val tysC = map gtyToTy tysG
	 val rtyC = case rtyG of SOME tyG => gtyToTy tyG
			       | NONE => UNITty
	 val tysC' = case tysC of [] => [UNITty,rtyC]
				| xs => xs@[rtyC]
     in
	 if m_public flags then
             SOME (VALdec((name,nowhere), tyListToArrowTy tysC', m_static flags))
	 else NONE
     end

fun fdeclToValdec {flags, name, ty, attrs} =
    let
         val tyC = (gtyToTy o jtyToTy) ty
     in
	 if f_public flags then
             SOME (VALdec((name,nowhere), tyC, f_static flags))
	 else NONE
    end


fun makeClassDec classfile =
    let
        val {flags, this, super, ifcs,
	     fdecls, mdecls, attrs, ...} = Decompile.toClassDecl classfile
	    handle (Io _) =>
		   error nowhere ("Io exception in ClassPath.makeClassDec")

         (* Could maybe modify toClassDecl so that it (optionally) doesn't
            decompile the bytecode:  this might speed things up. *)

	fun qualNameL c = (qualName c, nowhere)

        val this' = qualNameL this
        val super' = Option.map qualNameL super
        val ifcs' = List.map qualNameL ifcs
        val valdecs = (List.mapPartial fdeclToValdec fdecls) @
		      (List.mapPartial mdeclToValdec mdecls)
        val () = if hasFlag Classdecl.C_ACCinterface flags
		 then recordInterface (nameOf this')
		 else ()
    in
        CLASSdec(this', super', ifcs', valdecs)
    end


fun loadClassFile cname loc l =
    case l of
	[] => error loc ("Can't find "
			 ^ cname
			 ^  ".class in ["
			 ^ (listToString (fn x=>x) "," (classpath()))
			 ^ "]")
      | h::t =>
	case h of
	    DIR d =>
	    let
		val fullName = Util.makeFullFilename d cname "class"
	    in
		if Nonstdio.file_exists (fullName)
		then
		    makeClassDec (Classfile.inputClassFile fullName)
		else loadClassFile cname loc t
	    end
	  | ZIP z =>
	    let
		val fullName = Util.makeFullFilename "" cname "class"
	    in
		case ZipReader.inputMember (z, fullName) of
		    NONE => loadClassFile cname loc t
		  | SOME v => makeClassDec (Classfile.vectorToClass v)
	    end


(* getClassInfo takes a classname and load the appropriate classfile,
   recording names of superclass and interfaces.  Recursively loads
   superclass and interfaces (stopping when it finds a previously
   loaded class). Also makes class contexts for the classese loaded:
   these are saved in Context.sml and are available for use in Type.sml
   etc.   Returns a list of CLASSdecs for classes loaded. *)

(* Is it really necessary to go all the way up the hierarchy?  What we
   really need is to be sure that the class has been loaded before
   looking at superTable or interfaceTable;  we could simply check
   this before looking in the tables and load if necessary.

   What about method signatures and so on?
   Well,  there are calls to Type.resolveClass (= getClassInfo) in
   appropriate places in Type.sml. *)

fun getClassInfo (cname, loc) =
    if isKnownClass cname then () else
    let
	val () = vprint ("Loading " ^ cname ^ "\n")
	val class as CLASSdec (this, super, ifcs, fields) =
	    loadClassFile (cnameToFilename cname) loc (getClassHolders())
	val () = debugPrintln ("  superclass: " ^ (case super of NONE => "NONE" | SOME (c,_) => c))

    in
	markKnown cname;
	Context.makeClassContext class;

	Option.app (saveSuper cname) super;
	Option.app getClassInfo super;  (* OMIT ? *)

	saveInterfaces cname ifcs;
	app getClassInfo ifcs           (* OMIT ? *)
    end



(* Want to recursively load and update interfaces and superclasses,
   but not for internal classes *)

(* It's important here that we don't allow the user to define interfaces *)

fun saveUserClassInfo l =
    let
	fun doClassDec (cdec as CLASSdec((cname, loc), superOpt, intfs, meths)) =
	    let
		val () = debugPrintln ("saveUserClassInfo "^ cname)
		val () = markKnown cname
		val () = Context.makeClassContext cdec
		val () = saveInterfaces cname intfs
		val () = app getClassInfo intfs  (* OMIT ? *)
	    in
		case superOpt of
		    NONE => ()
		  | SOME super =>
		    if isInternal super then (* it'll be done anyway *)
			 saveSuper cname super
		    else
			(saveSuper cname super;
			 getClassInfo super)  (* OMIT ? *)
	    end
    in
	app markInternal l;
	app doClassDec l
    end


(* WATCH OUT: conceivably class and superclass could be in different directories *)

fun isSubclass classname supername =
    supername = "java.lang.Object"
    orelse classname = supername
    orelse
    let
	val () = getClassInfo (classname, nowhere)
    in
       	case Binarymap.peek(!superTable,classname) of
            SOME cl => cl = supername
		       orelse isSubclass cl supername
		       orelse isSubclassI classname supername
	  | NONE => isSubclassI classname supername
    end

and isSubclassI classname supername =
	case Binarymap.peek(!interfaceTable,classname) of
	    SOME intfs => List.exists (fn x => isSubclass x supername) intfs
	  | NONE => false

fun getSuperclass (class as (classname, loc)) =
    if classname = "java.lang.Object" then NONE
    else
	let
	    val () = getClassInfo class
	in
	    case Binarymap.peek(!superTable, classname) of
		SOME cl => SOME cl
              | NONE => SOME "java.lang.Object"
	end

fun getInterfaces (class as (classname, loc)) =
    let
	val () = getClassInfo class
    in
	case Binarymap.peek(!interfaceTable, classname) of
	SOME l => l
      | NONE => []
    end

fun isObjectType (OBJECTty _) = true
  | isObjectType (STRINGty)   = true
  | isObjectType _ = false

fun uniformObjectType (OBJECTty c) = (OBJECTty c)
  | uniformObjectType (STRINGty) = (OBJECTty "java.lang.String")
  | uniformObjectType x = x

fun typeIsSubclass t1 t2 =
    let
        val t1' = uniformObjectType t1
        val t2' = uniformObjectType t2
    in
        case (t1', t2') of
	    (OBJECTty c1, OBJECTty c2) => isSubclass c1 c2
          | _ => false
    end

end (* open *)
