(* Various kinds of context for use in Type.sml *)

(*

   A varCtx is a (string,Ty) dict.  We have one of these for the main class,
   which is carried about in Type.sml and (Mono.sml?)

   A classCtx is a (string, (Ty*instance) list) hashtable,  which maps
   (overloaded) identifiers to lists of types for fields/methods with that
   name.  There's one of these for each class which we encounter.

   There's a persistent table storing these in this file.  There's stuff
   in ClassPath.sml to ensure that each class is only entered in this
   table once

*)


local
    open Normsyn NAsyntfn Util
    structure P = Polyhash
    structure M = Splaymap
in

fun ctx_error s = Util.ierror ("[Context.sml]: " ^ s)


local
    val dataTypeInfo = ref (M.mkDict String.compare)
in

fun setDTinfo m = dataTypeInfo := m
fun getConInfo (cname, loc) =
    case M.peek (!dataTypeInfo, cname) of
	NONE => error loc ("[Context.sml]: Can't find constructor info for " ^ cname)
      | SOME h => h
end



fun typeType names ((TYPEdec(tvars, (tname,_), cons, loc), acc)) =
    let

	fun chkName loc x = if member x names then ()
				else error loc ("[Type]: Unknown type name " ^ x)

	fun checkTypeCon (TYPEcon(c, (tys,tloc), h, l)) =
	    let

		fun chk (ARROWty(a,b)) = (chk a; chk b)
		  | chk (ARRAYty a) = chk a
                  | chk (PRODUCTty tys) = app chk tys
		  | chk (CONty(tys, cc)) =  (app chk tys; chkName l cc)
		  | chk (TVARty v) =
		    if List.exists (fn (var,_) => var = #name v) tvars
		    then () else
		    error tloc
			  ("[Type]: Unknown type variable " ^ #name v ^ " in constructor")
		  | chk _ = ()

	    in
		app chk tys
	    end

	val () = app checkTypeCon cons   (* Should this be here?  Is Syncheck too early? *)

	val () = if cons = []
		 then error loc "Type must have at least one constructor"
		 else ()

	val tvars' = map (fn (x,_) => mkTvar x) tvars

	fun insertInfo (TYPEcon ((cname,_), (argtys,_), _, _), table) =
	    M.insert (table, cname, (CONty(tvars', tname), argtys))
    in
	List.foldl insertInfo acc cons
    end



(* TODO: check we don't clobber a type constructor name *)
(* Done in the parser? *)
fun saveDatatypeInfo types =
    let
	val names = map (fn TYPEdec(_,(n,_),_,_) => n) types
	val typeInfo = List.foldl (typeType names) (M.mkDict String.compare) types
    in
	setDTinfo typeInfo
    end


(* ---------------- Variable and class contexts ---------------- *)


type varCtx = (string, Ty) M.dict  (* instances, annotations? *)
(* Maps variables to types: no overloading allowed *)

fun ctxToString vctx =
    let
	val items = M.listItems vctx
	val s = listToString (fn (v,ty) => " " ^ v ^ " -> " ^ typeToString ty) "\n" items
    in
	s ^"\n"
    end

type classCtx = (string, (Ty * Instance) list) P.hash_table
(* Maps variables to lists of types/instances: overloading allowed *)

exception nameNotInClass

fun emptyCctx() = P.mkTable(P.hash,op=) (20, nameNotInClass)

local
    exception classContextNotFound

    val classInfo: (string, classCtx) P.hash_table ref =
	ref (P.mkTable (P.hash,op=) (20,classContextNotFound))
in
    fun resetClassInfo () =
	classInfo := P.mkTable(P.hash,op=)(20,classContextNotFound)

    fun addClassContext classname ctx =
	P.insert (!classInfo) (classname, ctx)

    fun getContext class =
	case P.peek (!classInfo) class of
	    NONE => ctx_error ("Can't find context for class " ^ class)
	  | SOME h => h

    fun cctxToDec (c,t) =
	let
	    val l = P.listItems t
	    fun mkValDec s (t,i) = VALdec ((s,nowhere), t, i)
	    val vdecs = List.concat (map (fn (s,h) => map (mkValDec s) h) l)
	in
	    CLASSdec ((c,nowhere), NONE, [], vdecs)  (* EEEEEEEEHHHH! *)
	end


    fun getClassDecs () = map cctxToDec (P.listItems (!classInfo))

    fun findClassCtx class =  P.peek (!classInfo) class

    fun addClassCtx (class,cctx) = P.insert (!classInfo) (class,cctx)

    fun resetClassInfo () = classInfo := P.mkTable(P.hash,op=)(20,classContextNotFound)

    fun printClassInfo () =
	let
	    val () = debugPrint "=== Class info===\n"
	    fun instToStr Absyn.STATIC = "S" | instToStr Absyn.INSTANCE = "I"
	    val l = P.listItems (!classInfo)

	    fun entry (t,i) = typeToString t ^ " (" ^ instToStr i ^ ")"
	    fun prCctx (c,t) =
		(debugPrint ("Context for class " ^ c ^ ":\n");
		 debugPrint (listToString (fn (s, l) => " " ^ s ^ " -> "
						   ^ (listToString entry ", " l)
				     ) "\n" (P.listItems t));
		 debugPrint "\n\n"
		)
	in
	    (app prCctx l; debugPrint "=====\n\n")
	end
end

fun insertInContext cctx (VALdec((x,u), ty, inst)) =
    let
	val l =
	    case P.peek cctx x of
		NONE => []
	      | SOME l => l
	val l' =
	    if member (ty,inst) l then l
	    else (ty,inst)::l
        (* Should we bother with inst?  Can a class have a static method
           with the same name and type as an instance method?  What about
           fields? *)
    in
	P.insert cctx (x, l')
    end

(* Adds a 0-argument initialiser to any class which has no <initialiser. *)
(* FIX: this is really fixing the bug that user-defined classes rely on the default
   initialiser without saying so. This should be fixed by allowing proper
   initialiser definition. *)

fun addInit methods =
    if List.exists (fn VALdec (("<init>",_), _, _) => true | _ => false) methods
    then methods
    else (VALdec(("<init>",nowhere), ARROWty(UNITty, UNITty), INSTANCE))::methods


fun makeClassContext (CLASSdec ((cname,u), super, interfaces, vals)) =
    let
	val cctx = emptyCctx ()
	val () = app (insertInContext cctx) (addInit vals)
    in
	addClassContext cname cctx
    end

fun getMember fmname cname loc =
    case findClassCtx cname of
	NONE => error loc ("Can't find info for class " ^ cname)
      | SOME ctx => P.peek ctx fmname

fun insertList (vctx, l) =
    List.foldl (fn ((a,b), c) => M.insert (c, a, b)) vctx l

fun valdecsToVarContext l =
    insertList (M.mkDict String.compare, map (fn VALdec ((name, u), ty, inst) => (name, ty)) l)

fun getbinding x (vctx: varCtx) loc =
    case M.peek (vctx, x) of
	NONE => error loc ("Variable " ^ x ^ " not in scope")
      | SOME h => h

fun getbindOpt x (vctx: varCtx) = M.peek (vctx, x)

fun addbinding ctx x ty = M.insert (ctx,x,ty)

val toList = M.listItems

end (* local structure ... *)
