(* kwxm: 29th October 2005 *)

local
    open Classdecl Jvmtype
    open GrailAbsyn GrailUtils

    structure P = Polyhash
    structure M = Splaymap

    (* Associate name of class with names of superclass/ interfaces *)
    val superTable = ref (Binarymap.mkDict String.compare)
    val interfaceTable = ref (Binarymap.mkDict String.compare)
    val flagTable = ref (Binarymap.mkDict String.compare)

    fun saveSuper C S = superTable := Binarymap.insert(!superTable, C, S)
    fun saveInterfaces C l = interfaceTable :=
	      Binarymap.insert(!interfaceTable,C, l)
    fun saveFlags C l = flagTable := Binarymap.insert(!flagTable, C, l)


    fun id x = x
    fun println s = print (s^"\n")
in

fun ctx_error s = ierror ("[Context]: " ^ s)

exception ClassError of string

(* -------------------- Info relating to methods and fields -------------------- *)


datatype Member =
	 FIELD  of string * Ty * Classdecl.field_access_flag list
       | METHOD of string * RTy * Ty list * Classdecl.method_access_flag list

datatype ClassInfo =
	 CLASSinfo of string
		      * string option
		      * string list
		      * Member list
		      * Classdecl.class_access_flag list

fun memName m =
    case m of FIELD (s,_,_) => s
	    | METHOD (s,_,_,_) => s

fun methFlags m =
    case m of FIELD _ => raise ClassError "Looking for method flags in field"
	    | METHOD (_,_,_,f) => f

fun fieldFlags m =
    case m of FIELD (_,_,f) => f
	    | METHOD _ => raise ClassError "Looking for field flags in method"

fun isField m =
    case m of FIELD  _ => true
	    | METHOD _ => false

fun isMethod m =
    case m of METHOD _ => true
	    | FIELD  _ => false

fun fullName C m =
    case m of 
	METHOD (name, rty, args, _) =>
	rtyToString rty ^ " " 
	^ C ^ "."
	^ name 
	^ "(" ^ listToString tyToString "," args ^ ")"
      | FIELD (name, ty, _) =>
	tyToString ty ^ " " 
	^ C ^ "."
	^ name 
	      
(* ---------------- Variable and class contexts ---------------- *)


type classCtx = (string, Member list) P.hash_table

(* Maps variables to lists of fields/methods: overloading allowed *)

val debugPrint = print

exception nameNotInClass

fun emptyCctx() = P.mkTable(P.hash,op=) (20, nameNotInClass)

local
    exception classContextNotFound

    val classInfo: (string, classCtx) P.hash_table ref =
	ref (P.mkTable (P.hash,op=) (20,classContextNotFound))
in
    fun resetClassInfo () =
	classInfo := P.mkTable(P.hash,op=)(20,classContextNotFound)

    fun addClassContext classname ctx =
	P.insert (!classInfo) (classname, ctx)

    fun getContext class =
	case P.peek (!classInfo) class of
	    NONE => ctx_error ("Can't find context for class " ^ class)
	  | SOME h => h

(*
    fun cctxToDec (c,t) =
	let
	    val l = P.listItems t
	    fun mkValDec s (t,i) = VALdec (s, t, i)
	    val vdecs = List.concat (map (fn (s,h) => map (mkValDec s) h) l)
	in
	    CLASSinfo ((c,nowhere), NONE, [], vdecs, [])  (* EEEEEEEEHHHH! *)
	end


    fun getClassDecs () = map cctxToDec (P.listItems (!classInfo))
*)

    fun findClassCtx class =  P.peek (!classInfo) class

    fun addClassCtx (class,cctx) = P.insert (!classInfo) (class,cctx)

    fun resetClassInfo () = classInfo := P.mkTable(P.hash,op=)(20,classContextNotFound)



    fun memToString m =
	case m of
	    FIELD (name, ty, flags) =>
	    "field " ^ name ^ ": "
	    ^ tyToString ty
	    ^ " [" ^ listToString fflagToString "|" flags ^"]"
	  | METHOD (name, rty, argtys, flags) =>
	    "method " ^ name ^ ": "
	    ^ rtyToString rty
	    ^ "(" ^ listToString tyToString "," argtys ^")"
	    ^ " [" ^ listToString mflagToString "|" flags ^"]"


    fun printClassInfo () =
	let
	    val () = debugPrint "=== Class info===\n"

	    val l = P.listItems (!classInfo)

	    fun prCctx (c,t) =
		(debugPrint ("Context for class " ^ c ^ ":\n");
		 debugPrint (listToString (fn (s, l) => " " ^ s ^ " -> "
							^ (listToString memToString ", " l)
					  ) "\n" (P.listItems t));
		 debugPrint "\n\n"
		)
	in
	    (app prCctx l; debugPrint "=====\n\n")
	end
end

fun insertInContext cctx m =
    let
	val name = memName m
	val l =
	    case P.peek cctx name of
		NONE => []
	      | SOME l => l
	val l' =
	    if member m l then l
	    else m::l
		 (* Can we have two methods with the same sig but diffferent flags?? *)
    in
	P.insert cctx (name, l')
    end


fun makeClassContext (CLASSinfo (cname, super, interfaces, members,flags)) =
    let
	val cctx = emptyCctx ()
	val () = app (insertInContext cctx) members
    in
	addClassContext cname cctx
    end

fun getMember fmname C =
    case findClassCtx C of
	NONE => exit ("Can't find info for class " ^ C)
      | SOME ctx => P.peek ctx fmname

fun getFields fname C = (* Get fields with name mname *)
    Option.map (List.filter isField) (getMember fname C)

fun getMethods mname C = (* Get methods with name mname *)
    Option.map (List.filter isMethod) (getMember mname C)

fun allMethods C =
    List.filter isMethod (List.concat (map snd (Polyhash.listItems (getContext C))))

fun sameSig rty args (* flags *) f =
    case f of
	METHOD(_,rty',args',flags') => 
	   rty' = rty 
	   andalso args' = args
(*	   andalso member M_ACCstatic flags' = member M_ACCstatic flags*)
      | _ => false

(* FIX:  Java regards two methods as being indistinguishable if they
   have the same name and argument types.  JVM can distiguish two
   such methods if they have different return types or if
   one is static and the other non-static. *)

(* Oops.  The verifier rejects a class if it has two methods which are 
   the same apart from the fact that one is static and the other is 
   non-static. It's OK to have a non-static method void m ()in a class C 
   and a static version in a superclass S: resolution fails if you do
      invokestatic <void C.m()> ()
   but succeeds for 
      invokestatic <void S.m()> ()

   static/non-static is important for resolution, but not for overriding.
*)

fun sameTy ty f =
    case f of
	FIELD(_,ty',_) => ty' = ty
      | _ => false

fun seekMethod m rty args C =
    case getMember m C of
	NONE => NONE
      | SOME l => List.find (sameSig rty args) l

fun seekField f ty C =
    case getMember f C of
	NONE => NONE
      | SOME l => List.find (sameTy ty) l

fun existsMethod m rty args C =
    case getMember m C of
	NONE => false
      | SOME l => List.exists (sameSig rty args) l


fun insertList (vctx, l) =
    List.foldl (fn ((a,b), c) => M.insert (c, a, b)) vctx l

fun getbinding x vctx =
    case M.peek (vctx, x) of
	NONE => exit ("Variable " ^ x ^ " not in scope")
      | SOME h => h

fun getbindOpt x vctx = M.peek (vctx, x)

fun addbinding ctx x ty = M.insert (ctx,x,ty)

val toList = M.listItems




(* -------------------- Loading classes -------------------- *)


fun printInterfaces () = Binarymap.app (fn (k,l) => println (k ^" --> ["
								 ^ listToString id ", " l
								 ^ "]")) (!interfaceTable)

fun cpError s = GrailUtils.ierror ("[GrailClasspath.sml]: " ^ s)

datatype classholder = DIR of string | ZIP of ZipReader.zipfile

local
    val cp: string list option ref  = ref NONE

    val classHolders = ref []
    val loaded = ref false

    fun setClassPath' s = (* s is a colon-separated list of directory names *)
    let
	val separator =
	    if member #";" (String.explode s)
	       orelse Path.validVolume {isAbs=true, vol="A:"}
	    then #";"      (* Windows *)
	    else #":"      (* Unix *)

	val files = String.tokens (fn c => c = separator) s
        (* Note that we haven't trimmed trailing slashes; we always
           manipulate paths and filenames using the functions in Path,
           and these make sure that slashes (or other separators)
           are in the correct places *)
    in
	cp := SOME files
    end

    fun initClasspath () =
	case (!cp) of
	    NONE => (* not set from command-line, so look elsewhere *)
	    let
		val pwd = getOpt (Process.getEnv "PWD", ".")
		val default =
		    case Process.getEnv "GRAIL_CLASSPATH"
		     of NONE => (* try ${HOME}/classes *)
			let in
			    case Process.getEnv "HOME"
			     of NONE => pwd
			      | SOME home =>
				let val f = Path.joinDirFile
						{dir=home, file="classes"}
				in if Nonstdio.file_exists f then f ^ ":" ^ pwd else pwd end
			end
		      | SOME s => s
	    in
		setClassPath' default
	    end
	  | SOME p => ()
in

fun classpath () = valOf (!cp)
fun setClassPath s = setClassPath' s


(* The next function can be quite time consuming,  so we don't want to call
   it unless we're definitely using OO stuff;  this is why the stuff with
   'loaded' is there. *)

fun getClassHolders () =
    if (!loaded) then !classHolders
    else
    let
	val () = loaded := true
	val () = initClasspath ()
	fun doFile f =
	    case (#ext (Path.splitBaseExt f)) of
		SOME "jar" => ZIP (ZipReader.open_in f)
	      |	SOME "zip" => ZIP (ZipReader.open_in f)
	      | _ => DIR f
    in
	classHolders := map doFile (valOf (!cp));
	!classHolders
    end

fun closeJars () = app (fn ZIP f => ZipReader.close_in f | _ => ()) (!classHolders)
end  (* local *)


local
    val knownClasses = ref (Binaryset.empty String.compare)
    val internalClasses = ref (Binaryset.empty String.compare)  (* Classes defined in current file *)
    val interfaceTable = ref (Splayset.empty String.compare)
    (* Table of names of "classes" which are really interfaces *)
in

fun markKnown c = knownClasses := Binaryset.add(!knownClasses,c)
fun isKnownClass c = Binaryset.member(!knownClasses,c)

fun markInternal (CDEF(_,cname,_,_,_,_,_)) =
    internalClasses := Binaryset.add(!internalClasses,cname)
fun isInternal c = Binaryset.member(!internalClasses,c)

fun recordInterface c = (interfaceTable := Splayset.add(!interfaceTable,c))
fun isInterface c = Splayset.member(!interfaceTable,c)  (* Needed in Phi.sml *)

end

fun cnameToFilename cname = String.translate (fn #"." => "/" | x => str x) cname
fun filenameToClass fname = String.translate (fn #"/" => "." | x => str x) fname

fun f_public flags = member Classdecl.F_ACCpublic flags
fun f_static flags = member Classdecl.F_ACCstatic flags
fun m_public flags = member Classdecl.M_ACCpublic flags
fun m_static flags = member Classdecl.M_ACCstatic flags

fun fdefToInfo (FDEF(flags, ty, name,_)) =
    FIELD (name, ty, flags)

fun mdefToInfo (MDEF(flags, rty, name, args, _)) =
    METHOD (name, rty, map fst args, flags)

fun cdefToClassDec (CDEF(flags, cname, superOpt, intfs, fields, methods, _)) =
    CLASSinfo (cname, superOpt, intfs, (map fdefToInfo fields)@(map mdefToInfo methods), flags)


fun qualName C =
    let
        fun f (a::b::t) = f ((a ^"."^b)::t)
          | f [x] = x
          | f [] = ""
        val toks = (Jvmtype.packages C)@[Jvmtype.className C]
    in f toks end

fun jtyToTy t =
    case t of
	Tint      => GrailAbsyn.INTty
      | Tboolean  => GrailAbsyn.BOOLEANty
      | Tfloat    => GrailAbsyn.FLOATty
      | Tclass c  => GrailAbsyn.REFty (qualName c)
      | Tarray ty => GrailAbsyn.ARRAYty (jtyToTy ty)
      | Tchar     => GrailAbsyn.CHARty
      | Tbyte     => GrailAbsyn.BYTEty
      | Tshort    => GrailAbsyn.SHORTty
      | Tlong     => GrailAbsyn.LONGty
      | Tdouble   => GrailAbsyn.DOUBLEty

fun jtyToRTy (SOME t) = SOME (jtyToTy t)
  | jtyToRTy NONE = NONE


fun fdeclToInfo {flags, name, ty, attrs} =
    FIELD (name, jtyToTy ty, flags)


fun mdeclToInfo {flags, name, msig = (tys, rty), attrs} =
    METHOD (name, jtyToRTy rty, map jtyToTy tys, flags)


fun makeClassDec classfile =
    let
        val {flags, this, super, ifcs,
	     fdecls, mdecls, attrs, ...} = Decompile.toClassDecl classfile
	    handle (Io _) =>
		   exit ("Io exception in GrailClasspath.makeClassDec")

         (* Could maybe modify toClassDecl so that it (optionally) doesn't
            decompile the bytecode:  this might speed things up. *)


        val this' = qualName this
        val super' = Option.map qualName super
        val ifcs' = List.map qualName ifcs
        val members = (map fdeclToInfo fdecls) @ (map mdeclToInfo mdecls)
        val () = if member Classdecl.C_ACCinterface flags
		 then recordInterface (this')
		 else ()
    in
        CLASSinfo(this', super', ifcs', members, flags)
    end


fun loadClassFile C l =
    case l of
	[] => exit ("Can't find "
		    ^ C
		    ^  ".class in ["
		    ^ (listToString (fn x=>x) "," (classpath()))
		    ^ "]")
      | h::t =>
	case h of
	    DIR d =>
	    let
		val fullName = GrailUtils.makeFullFilename d C "class"
	    in
		if Nonstdio.file_exists (fullName)
		then
		    makeClassDec (Classfile.inputClassFile fullName)
		else loadClassFile C t
	    end
	  | ZIP z =>
	    let
		val fullName = GrailUtils.makeFullFilename "" C "class"
	    in
		case ZipReader.inputMember (z, fullName) of
		    NONE => loadClassFile C t
		  | SOME v => makeClassDec (Classfile.vectorToClass v)
	    end


(* resolveClass takes a classname and load the appropriate classfile,
   recording names of superclass and interfaces.  Recursively loads
   superclass and interfaces (stopping when it finds a previously
   loaded class). Also makes class contexts for the classese loaded:
   these are saved in Context.sml and are available for use in Type.sml
   etc.   Returns a list of CLASSinfos for classes loaded. *)

(* Is it really necessary to go all the way up the hierarchy?  What we
   really need is to be sure that the class has been loaded before
   looking at superTable or interfaceTable;  we could simply check
   this before looking in the tables and load if necessary.

   What about method signatures and so on?
   Well,  there are calls to Type.resolveClass (= resolveClass) in
   appropriate places in Type.sml. *)

fun resolveClass C =  (* resolve class and all superclasses and superinterfaces *)
let
    fun f done C =
	if member C done then raise ClassError ("Circularity error for class " ^ C)
	else
	    if isKnownClass C then () else
	    let
		val () = vprint ("Loading " ^ C ^ "\n")

		val class as CLASSinfo (this, super, ifcs, fields, flags) =
		    loadClassFile (cnameToFilename C) (getClassHolders())
	    in
		markKnown C;
		makeClassContext class;

		Option.app (saveSuper C) super;
		Option.app (f (C::done)) super;

		saveInterfaces C ifcs;
		app resolveClass ifcs;

		saveFlags C flags
	    end
in
    f [] C
end



(* Want to recursively load and update interfaces and superclasses,
   but not for internal classes *)

(* It's important here that we don't allow the user to define interfaces *)
(* ---  er.... *)


fun saveUserClassInfo l (* CDEF list *) =
    let
	fun doClassDec (cdef as CDEF(flags, C, superOpt, intfs, _,_,_)) =
	    let
(*		val () = println ("saveUserClassInfo "^ C)*)
		val () = markKnown C
		val () = makeClassContext (cdefToClassDec cdef)
		val () = saveInterfaces C intfs
		val () = app resolveClass intfs  (* OMIT ? *)
		val () = saveFlags C flags
	    in
		case superOpt of
		    NONE => ()
		  | SOME S => (* We check in Compile.sml that S <> C *)
		    if isInternal S then (* it'll be done anyway *)
		         saveSuper C S
		    else
			(saveSuper C S;
			 resolveClass S)  (* OMIT ? *)
	    end
    in
	app markInternal l;
	app doClassDec l
    end


(* WATCH OUT: conceivably class and superclass could be in different directories *)

fun isSubclass C D =
    D = "java.lang.Object"
    orelse C = D
    orelse
    let
	val () = resolveClass C
    in
       	case Binarymap.peek(!superTable,C) of
            SOME S => S = D
		      orelse isSubclass S D
		      orelse isSubclassI C D
	  | NONE => isSubclassI C D
    end

and isSubclassI C D =
	case Binarymap.peek(!interfaceTable,C) of
	    SOME intfs => List.exists (fn x => isSubclass x D) intfs
	  | NONE => false

fun getSuperclass C =
    if C = "java.lang.Object" then NONE
    else
	let
	    val () = resolveClass C
	in
	    case Binarymap.peek(!superTable, C) of
		SOME cl => SOME cl
              | NONE => SOME "java.lang.Object"
	end

fun getInterfaces C =
    let
	val () = resolveClass C
    in
	case Binarymap.peek(!interfaceTable, C) of
	SOME l => l
      | NONE => []
    end

fun implements C I =
    let
	val l = getInterfaces C
    in
	member I l  (* FIX: Do we have to check the superinterfaces as well? *)
    end

fun getCFlags C =
    let
	val () = resolveClass C
    in
	case Binarymap.peek(!flagTable, C) of
	SOME l => l
      | NONE => []
    end

fun isObjectType (REFty _) = true  (* Don't need this *)
  | isObjectType _ = false

fun uniformObjectType (REFty c) = (REFty c)
  | uniformObjectType x = x

fun typeIsSubclass t1 t2 =
    let
        val t1' = uniformObjectType t1
        val t2' = uniformObjectType t2
    in
        case (t1', t2') of
	    (REFty c1, REFty c2) => isSubclass c1 c2
          | _ => false
    end

end (* open *)
