(* Compile Grail to JVML.
   Careful: this file is shared by Camelot and gdf via symbolic links.
   The master copy is progs/Grail/gdf/src/Compile.sml *)

val testing = ref false
val setACCsuper = ref true

open Bytecode
open Jvmtype

open GrailAbsyn
open GrailUtils
open Persistent

val debug = ref false

fun pr x = TextIO.print (x^" ")
fun prl x = TextIO.print (x^"\n")

fun pr x = ()
fun prl x = ()

fun println x = TextIO.print (x^"\n")

(* Don't want messages mixed with syntax going to stdout *)
(*
fun printToStdErr s =
    if not (!ToyGrailAbsyn.shut_up) then
	TextIO.output(TextIO.stdErr, s)
else ()
*)
fun printToStdErr s = TextIO.output(TextIO.stdErr, s)

val typeOfValue = Typecheck.typeOfValue
val typeOfVar = Typecheck.typeOfVar

(* -------------------------------- Bytecode emission -------------------------------- *)

fun mdescToMethodref (MDESC(ty, name, argtypes)) =
    let
	val l = String.tokens (fn c => (c = #".")) name
	val n = List.length l
	val msig = (map tyToJty argtypes, rtyToJty ty)
	val name = List.nth(l,n-1)
	val (pkgs, cname) =
	    if n=1 then
		(TextIO.print ("WARNING: suspect method descriptor (" ^ name ^ ")\n");
		 ([],"")
		 )
	    else (List.take(l,n-2), List.nth(l,n-2))
    in
	{name=name, msig=msig, class = Jvmtype.class{pkgs=pkgs, name=cname}}
    end

fun mdescToDynamicref (MDESC(ty, name, argtypes)) =  (* for invokedynamic: no class *)
    let
	val () = if Char.contains name #"." then
		     localError ("invokeinterface: " ^ name ^ " contains '.'")
		     else ()
	val msig = (map tyToJty argtypes, rtyToJty ty)
    in
	{name=name, msig=msig}
    end


fun fdescToFieldRef (FDESC(fty,name)) =
  let val l = String.tokens (fn c => (c = #".")) name
	val n = List.length l
	val ty = tyToJty fty
	val name = List.nth(l,n-1)
	val (pkgs, cname) =
	    if n=1 then
		(TextIO.print ("WARNING: suspect field descriptor (" ^ name ^ ")\n");
		 ([],"")
		 )
	    else (List.take(l,n-2), List.nth(l,n-2))
	(* val _ =BasicIO.print ("cname = "^cname^", field = "^ name ^"\n") *)
  in
      {name=name, ty=ty, class = class{pkgs=pkgs, name=cname}}
  end


fun loadVar v A = (
   case lookupVar v  of   (* should really use findVar here *)
     NONE => localError ("Eh? can't find variable " ^ v)
   | SOME (t,l) =>
     let in
	 case t of
	     BOOLEANty => (Jiload l):: A
           (* Probably never happens: t will have been converted to INTty
              by the call to typeOfPrimop in declareVar *)
           | BYTEty    => (Jiload l):: A
	   | SHORTty   => (Jlload l):: A
           | INTty     => (Jiload l):: A
	   | LONGty    => (Jlload l):: A
	   | CHARty    => (Jiload l):: A
	   | FLOATty   => (Jfload l):: A
	   | DOUBLEty  => (Jdload l):: A
	   | REFty _   => (Jaload l):: A
	   | ARRAYty _ => (Jaload l):: A
     end
)


fun storeVar v A = (
   case lookupVar v of   (* should really use findVar here *)
     NONE => localError ("Eh? can't find variable " ^ v)
   | SOME (t,l) =>
     let in
	 case t of
             BOOLEANty => (Jistore l) :: A
	   | BYTEty => (Jistore l) :: A
	   | SHORTty => localError "Unexpected SHORTty in storeVar"
	   | INTty => (Jistore l) :: A
	   | LONGty => (Jlstore l) :: A
	   | CHARty => localError "Unexpected CHARty in storeVar"
           | FLOATty => (Jfstore l) :: A
	   | DOUBLEty => (Jdstore l) :: A
	   | REFty _ => (Jastore l) :: A
	   | ARRAYty _ => (Jastore l) :: A
     end
)

local
    fun loadInt n A =   (Jiconst (Int32.fromInt n)) :: A
in
fun loadVal w A =
    case w of VARval v    => loadVar v A
	    | BYTEval b   => loadInt b A
	    | SHORTval s  => loadInt s A
	    | INTval i    => loadInt i A  (* FIX: May be too short *)
	    | LONGval l   => (Jlconst l)::A
	    | CHARval c   => loadInt c A
	    | FLOATval r  => (Jfconst r)::A
	    | DOUBLEval d => (Jdconst d)::A
	    | STRINGval s =>  (* quotes should have been removed by now *)
	      let in
		  case String.fromString s of
		      SOME x => (Jsconst x) :: A
		    | NONE => localError ("Error in string literal \"" ^ s ^ "\"")
	      end
	    | CLASSval c => (Jclassconst (CLASS(qualNameToClass c))) :: A
	    | NULLval (_,t) => (* Save type marker for conversion into metadata *)
	      let
		  val () = Metadata.saveMarker t
	      in
		  (Jaconst_null) :: A
	      end
end

fun loadArgs [] A = A
  | loadArgs (h::t) A = loadVar h (loadArgs t A)

fun loadArgVals [] A = A
  | loadArgVals (h::t) A = loadVal h (loadArgVals t A)

fun storeArgs l A =
let
    fun storeArgs' [] A = A
      | storeArgs' ((ty,name)::t) A = storeVar name (storeArgs' t A)
in
    storeArgs' (List.rev l) A
end


fun compileCast c =
    case c of
	dtof => Jd2f
      | dtoi => Jd2i
      | dtol => Jd2l
      | ftod => Jf2d
      | ftoi => Jf2i
      | ftol => Jf2l
      | itob => Ji2b
      | itoc => Ji2c
      | itod => Ji2d
      | itof => Ji2f
      | itol => Ji2l
      | itos => Ji2s
      | ltod => Jl2d
      | ltof => Jl2f
      | ltoi => Jl2i


(* Some duplication here of checks carried out during typechecking *)

fun compilePrimOp p A =
    case p of
	VALop v => loadVal v A
      | UNARYop (oper, v) =>
	let
	    val t = typeOfValue v
	    val opcode =
		case (oper, t) of
		    (NEGop, INTty)    => Jineg
		  | (NEGop, LONGty)   => Jlneg
		  | (NEGop, FLOATty)  => Jfneg
		  | (NEGop, DOUBLEty) => Jdneg
		  | _ => primOpErr p  ("Can't apply " ^ unaryOpToString oper
				       ^ " to " ^ tyToString t)
	in
	    loadVal v (opcode::A)
	end

      | IINCop (v,n) =>
	let in case
		lookupVar v of
		NONE => localError "?: iinc"
	      | SOME (t,l) => if t = INTty
			      then Jiinc {var = l, const = n} :: A
			      else localError ("iinc applied to " ^ tyToString t)
	end

      | BINop (oper, a, b) =>
	let
	    val t = typeOfValue a
	in
	    if isIntegral t then  (* FIX: OK for byte/short/char??? *)
	    let
		val opcode =
		    case oper of
			ADDop  => Jiadd
		      | SUBop  => Jisub
		      | MULop  => Jimul
		      | DIVop  => Jidiv
		      | MODop  => Jirem
		      | CMPop  => Jisub
		      | CMPGop => primOpErr p "Operator only applies to float/double"
		      | CMPLop => primOpErr p "Operator only applies to float/double"
		      | ANDop  => Jiand
		      | ORop   => Jior
		      | XORop  => Jixor
		      | SHLop  => Jishl
		      | SHRop  => Jishr
		      | USHRop => Jiushr
	    in
		loadVal a (loadVal b (opcode::A))
	    end
	    else
		case t of
		    LONGty =>
		    let
			val opcode =
			    case oper of
				ADDop  => Jladd
			      | SUBop  => Jlsub
			      | MULop  => Jlmul
			      | DIVop  => Jldiv
			      | MODop  => Jlrem
			      | CMPop  => Jlcmp
			      | CMPLop => primOpErr p "Operator only applies to float/double"
			      | CMPGop => primOpErr p "Operator only applies to float/double"
			      | ANDop  => Jland
			      | ORop   => Jlor
			      | XORop  => Jlxor
			      | SHLop  => Jlshl
			      | SHRop  => Jlshr
			      | USHRop => Jlushr
		    in
			loadVal a (loadVal b (opcode:: A))
		    end

		  | FLOATty =>
		    let
			val opcode =
			    case oper of
				ADDop  => Jfadd
			      | SUBop  => Jfsub
			      | MULop  => Jfmul
			      | DIVop  => Jfdiv
			      | MODop  => Jfrem
			      | CMPop  => primOpErr p "Ambiguous comparison: use cmpg or cmpl for floats"
			      | CMPGop => Jfcmpg
			      | CMPLop => Jfcmpl
			      | _ => localError ("No '" ^ binOpToString oper ^ "' operator for floats")
		    in
			loadVal a (loadVal b (opcode:: A))
		    end

		  | DOUBLEty =>
		    let
			val opcode =
			    case oper of
				ADDop  => Jdadd
			      | SUBop  => Jdsub
			      | MULop  => Jdmul
			      | DIVop  => Jddiv
			      | MODop  => Jdrem
			      | CMPop  => primOpErr p "Ambiguous comparison: use cmpg or cmpl for doubles"
			      | CMPGop => Jdcmpg
			      | CMPLop => Jdcmpl
			      | _ => localError ("No '" ^ binOpToString oper ^ "' operator for doubles")
		    in
			loadVal a (loadVal b (opcode:: A))
		    end

		  | _ => localError ("Type " ^ tyToString t ^ " found in binop")
	end

      | NEWop(MDESC(r,s,tl), vs) =>
	(Jnew (qnameToJClass s)) :: Jdup :: (loadArgVals vs
		(Jinvokespecial (mdescToMethodref (MDESC(r, s ^ ".<init>", tl))) :: A))
      | CHECKCASTop (t,v) =>
	loadVar v (Jcheckcast(tyToClassRef t) :: A)
      | INSTANCEop  (t,v) =>
        loadVar v (Jinstanceof(tyToClassRef t) :: A)
      | INVOKESTATICop (md, args) =>
        loadArgVals(args) (Jinvokestatic (mdescToMethodref md) :: A)
      | INVOKEDYNAMICop (inst, md, args) =>
	let val (_,loc) = findVar inst in
            (Jaload loc) :: (loadArgVals args (Jinvokedynamic (mdescToDynamicref md) :: A))
	end
      | INVOKEINTERFACEop (inst, md, args) =>
	let val (_,loc) = findVar inst in
            (Jaload loc) :: (loadArgVals args (Jinvokeinterface (mdescToMethodref md) :: A))
	end
      | INVOKESPECIALop (inst, md, args) =>
	let val (_,loc) = findVar inst in
            (Jaload loc) :: (loadArgVals args (Jinvokespecial (mdescToMethodref md) :: A))
	end
      | INVOKEVIRTUALop (inst, md, args) =>
	let val (_,loc) = findVar inst in
            (Jaload loc) :: (loadArgVals args (Jinvokevirtual (mdescToMethodref md) :: A))
	end
      | GETFIELDop (inst ,fd) =>
	let val (_,loc) = findVar inst in
	    (Jaload loc) :: (Jgetfield (fdescToFieldRef fd) :: A)
	end
      | PUTFIELDop (inst, fd, input) =>
	let val (_,loc) = findVar inst in
	    (Jaload loc) :: (loadVal input (Jputfield (fdescToFieldRef fd) :: A))
	end
      | GETSTATICop fd => (Jgetstatic (fdescToFieldRef fd))::A
      | PUTSTATICop (fd, input) =>
	loadVal input (Jputstatic (fdescToFieldRef fd) :: A)

      | GETop(aa,i) =>
	let
	    val t = typeOfValue aa
	in
	    case t of
		ARRAYty INTty      => loadVal aa (loadVal i (Jiaload :: A))
	      | ARRAYty BOOLEANty  => loadVal aa (loadVal i (Jbaload :: A))  (* FIX: ?? *)
	      | ARRAYty BYTEty     => loadVal aa (loadVal i (Jbaload :: A))
	      | ARRAYty CHARty     => loadVal aa (loadVal i (Jcaload :: A))
	      | ARRAYty SHORTty    => loadVal aa (loadVal i (Jsaload :: A))
	      | ARRAYty LONGty     => loadVal aa (loadVal i (Jlaload :: A))
	      | ARRAYty FLOATty    => loadVal aa (loadVal i (Jfaload :: A))
	      | ARRAYty DOUBLEty   => loadVal aa (loadVal i (Jdaload :: A))
	      | ARRAYty(REFty _)   => loadVal aa (loadVal i (Jaaload :: A))
	      | ARRAYty(ARRAYty _) => loadVal aa (loadVal i (Jaaload :: A))
	      | _ => primOpErr p ("Unexpected type " ^ tyToString t)
	end

      | SETop(aa,i,v) => (* Changed to return no value = no dup *)
	let
	    val t = typeOfValue aa
	in
	    case t of
		ARRAYty INTty      => loadVal aa (loadVal i (loadVal v (Jiastore :: A)))
	      | ARRAYty BOOLEANty  => loadVal aa (loadVal i (loadVal v (Jbastore :: A)))
	      | ARRAYty BYTEty     => loadVal aa (loadVal i (loadVal v (Jbastore :: A)))
	      | ARRAYty CHARty     => loadVal aa (loadVal i (loadVal v (Jcastore :: A)))
	      | ARRAYty SHORTty    => loadVal aa (loadVal i (loadVal v (Jsastore :: A)))
	      | ARRAYty LONGty     => loadVal aa (loadVal i (loadVal v (Jlastore :: A)))
	      | ARRAYty FLOATty    => loadVal aa (loadVal i (loadVal v (Jfastore :: A)))
	      | ARRAYty DOUBLEty   => loadVal aa (loadVal i (loadVal v (Jdastore :: A)))
	      | ARRAYty (REFty _)  => loadVal aa (loadVal i (loadVal v (Jaastore :: A)))
	      | ARRAYty(ARRAYty _) => loadVal aa (loadVal i (loadVal v (Jaastore :: A)))
	      | _ => primOpErr p ("Unexpected type " ^ tyToString t)
	end

      | LENGTHop(aa) => loadVar aa (Jarraylength :: A)

      | NEWARRAYop(t,l) =>
	let
	    val dim = length l
	    fun fixDim ty =
		let fun f ty n =
			if n=1 then ty
			else f (ARRAYty ty) (n-1)
		in
		    f t dim
		end

(*	    val () = print ("Array type -> " ^ tyToString t ^ "\n")*)
	in
	    loadArgVals l ((Jnewarray { elem = tyToJty (fixDim t), dim=dim}) :: A)
	end
(* multinewarray contains type of resulting array:  if we
   make a 2-dim array of int[][][] then the opcode contains int[][][][][].
   Do we want to do this or not?? *)
(* We do *)


      | CASTop (c,v) => loadVal v (compileCast c :: A)
      | LOCKop v => loadVal v (Jmonitorenter :: A)
      | UNLOCKop v => loadVal v (Jmonitorexit :: A)

      | THROWop v =>
	let
	    val t = typeOfVar v
	in
	    case t of
		REFty _ => loadVar v (Jathrow :: A)
	      | _ => localError ("Trying to throw value of type " ^ tyToString t)
              (* FIX: Should really check that obj is Throwable *)
	end

fun compileLetDec (VALdec (v,p)) A = compilePrimOp p (storeVar v A)
  | compileLetDec (VOIDdec p) A = compilePrimOp p A

fun compileLetDecs letdecs A =
    let fun comp [] A = A
	  | comp (h::t) A = compileLetDec h (comp t A)
    in
	comp letdecs A
end
(* This could be more tail-recursive; we don't usually have thousands
   of letDecs,  so it's probably fairly safe. *)


fun compilePrimRes (OPres p) A = compilePrimOp p (Jreturn :: A)
  | compilePrimRes VOIDres  A = Jreturn :: A
  | compilePrimRes (FUNres (fname, params)) A =
    let val (_,_,dest) = findFun fname
    in
	(Jgoto dest) :: A
    end

fun compileComparison ty test dest A =
    if isIntegral ty then
	case test of
	    EQcmp => (Jif_icmpeq dest) :: A
	  | NEcmp => (Jif_icmpne dest) :: A
	  | LTcmp => (Jif_icmplt dest) :: A
	  | LEcmp => (Jif_icmple dest) :: A
	  | GTcmp => (Jif_icmpgt dest) :: A
	  | GEcmp => (Jif_icmpge dest) :: A

    else if isReference ty then
	case test of
	    EQcmp => (Jif_acmpeq dest) :: A
	  | NEcmp => (Jif_acmpne dest) :: A
	  | _ => localError ("No '" ^ cmpToString test ^ "' test for references")

    else
	let
	    val cmp =
		case ty of
		    LONGty   => Jlcmp
		  | FLOATty  => Jfcmpg (* FIX: what about fcmpg? *)
		  | DOUBLEty => Jdcmpg (* Probably shouldn't allow comps for floats:
					  g/l ambiguity (javac produces cmpg) *)
		  | _ => localError ("Attempting comparison on type " ^ tyToString ty )
	in
	    case test of
		EQcmp => cmp :: (Jifeq dest) :: A
	      | NEcmp => cmp :: (Jifne dest) :: A
	      | LTcmp => cmp :: (Jiflt dest) :: A
	      | LEcmp => cmp :: (Jifle dest) :: A
	      | GTcmp => cmp :: (Jifgt dest) :: A
	      | GEcmp => cmp :: (Jifge dest) :: A
	end

(*
    else case ty of
	     LONGty =>
	     let in
		 case test of
		     EQcmp => Jlcmp :: (Jifeq dest) :: A
		   | NEcmp => Jlcmp :: (Jifne dest) :: A
		   | LTcmp => Jlcmp :: (Jiflt dest) :: A
		   | LEcmp => Jlcmp :: (Jifle dest) :: A
		   | GTcmp => Jlcmp :: (Jifgt dest) :: A
		   | GEcmp => Jlcmp :: (Jifge dest) :: A
	     end
	   | _ => localError ("Attempting comparison on type " ^ tyToString ty )
*)
fun compileIntTest v tst dest A =
    case tst of
	EQ0 => (Jifeq dest) :: A
      | NE0 => (Jifne dest) :: A
      | LT0 => (Jiflt dest) :: A
      | LE0 => (Jifle dest) :: A
      | GT0 => (Jifgt dest) :: A
      | GE0 => (Jifge dest) :: A
      | _ => localError (testToString tst ^ " applied to " ^ Typecheck.valAndType v)

(* FIX: should this type error be in Typecheck.sml? *)


fun compileRefTest v tst dest A =
    case tst of
	EQNULL => (Jifnull dest) :: A
      | NENULL => (Jifnonnull dest) :: A
      | _ => localError (testToString tst ^ " applied to " ^ Typecheck.valAndType v)

(* FIX: should this type error be in Typecheck.sml? *)

fun compileCaseRes (low, high, cases, (dfname,_)) A =
    let
	fun doCases c acc =
	    case c of
		[] => rev acc
	      | (_,fname,_)::t => let
		    val (_,_,dest) = findFun fname
		in
		    doCases t (dest::acc)
		end

	val labels = doCases cases []
	val (_,_,ddest) = findFun dfname
    in
	Jtableswitch {default=ddest,
		      offset = Int32.fromInt low,
		      targets=Vector.fromList labels} :: A
    end


fun compileLookupRes (cases, (dfname,_)) A =
    let
	fun doCases c acc =
	    case c of
		[] => rev acc
	      | (n,fname,_)::t => let
		    val (_,_,dest) = findFun fname
		in
		    doCases t ((Int32.fromInt n,dest)::acc)
		end

	val targets = doCases cases []
	val (_,_,ddest) = findFun dfname
    in
	Jlookupswitch {default=ddest,
		       cases=targets} :: A
    end

fun compileResult r A =
    case r of
	PRIMres p => compilePrimRes p A
      | CHOICEres (v, cmp, w, p1, p2) =>
	let val dest1 = newLabel()
	in
	    loadVal v (
	    loadVal w (
            compileComparison (typeOfValue v) cmp dest1 (
	    compilePrimRes p2 ((Jlabel dest1) :: (compilePrimRes p1 A)))))
	end
      | TESTres (tst, v, p1, p2) =>
	let
	    val dest1 = newLabel()
	    val A' = compilePrimRes p2 ((Jlabel dest1) :: (compilePrimRes p1 A))

	in
	    if isReference (typeOfValue v)
	    then loadVal v (compileRefTest v tst dest1 A')
	    else loadVal v (compileIntTest v tst dest1 A')
	end
      | CASEres (v, low, high, cases, default) =>
	loadVar v (compileCaseRes (low, high, cases, default) A)
      | LOOKUPres (v, cases, default) =>
	loadVar v (compileLookupRes (cases, default) A)


fun compileFunBody (FUNbody(letdecs, result)) A
		       = compileLetDecs letdecs (compileResult result A)

(* Now we're compiling a fun declaration.  At this point,
   the name of the function and its local variables have already
   been declared *)

fun compileFunDec (FDEC(fname, args, body)) A =
    let
	val () = setCurrentFunction fname
	val (_,_,entry) = findFun fname
    in
	Jlabel(entry) :: (compileFunBody body A)
    end

fun compileFunDecs l A =
    let fun
	compileFuns [] A = A
      | compileFuns (h::t) A = compileFunDec h (compileFuns t A)
    in compileFuns l A end
(* Again,  could be more tail-recursive.  If this gets changed, watch out
   that it doesn't mess up the order of the metadata null-markers. *)

fun compileMethodBody mbody =
    case mbody of
	MBODY(letdecs,fundecs,result) =>
	Jlabel(!codeStart) :: (
        compileLetDecs letdecs (
	compileResult result (
	compileFunDecs fundecs [Jlabel(!codeEnd)])))
      | NOBODY => []

fun compile_mdef (mdef as MDEF(flags, rettype, mname, params, mbody)) =
    let
	val () = Typecheck.typeCheckMdef mdef
    in
    case mbody of
	MBODY (_,fundecs,_) =>
	let
	    val () = Metadata.resetMarkers ()  (* Camelot type markers for null values *)
	    val () = setCurrentMethod mname
	    val () = setCurrentFunction ""

	    val () =
		if !testing then
		    let
			val () = print ("(*---- " ^ mname ^" ----*)\n")
			val () = FlowGraph.makeDefs mbody
		    in
			()
		    end
		else ()

	    val code = compileMethodBody mbody
	    val attrs = [CODE{stack=Stackdepth.maxdepth code [],
			      locals=numLocals(),
			      code=code, hdls=[],
			      attrs=[getLocals()]},
			 Metadata.makeFunInfo findVar (getFunLabels()),
			 Metadata.makeNullInfo()
			]
	    val msig =(map (fn (t,v) => tyToJty t) params, rtyToJty rettype)

	  (*	val _ = TextIO.print (Int.toString(numLocals ()) ^ " local variables\n");*)
	  (*	val _ = dumpLocals();*)

	in
	    {flags=flags, name=mname, msig=msig, attrs=attrs}
	end

      | NOBODY =>
	let
	    val () = setCurrentMethod mname
	    val () = setCurrentFunction ""
	    val msig =(map (fn (t,v) => tyToJty t) params, rtyToJty rettype)

	in
	    {flags=flags, name=mname, msig=msig, attrs=[]}
	end
    end

fun compileConstVal v =
    case v of
	 VARval _    => fieldErr "Bad initialiser: need constant"
       | BYTEval n   => Cint (Int32.fromInt n)
       | SHORTval n  => Cint (Int32.fromInt n)
       | INTval n    => Cint (Int32.fromInt n)
       | CHARval n   => Cint (Int32.fromInt n)
       | LONGval l   => Clong l
       | FLOATval f  => Cfloat f
       | DOUBLEval d => Cdouble d
       | STRINGval s => Cstring s
       | CLASSval _  => fieldErr "Bad initialiser: unexpected class constant"
       | NULLval _   => Cnull

fun compile_fdef (fdef as FDEF(flags,ty,name,fval)) =
    let
	val () = Typecheck.typeCheckFdef fdef
	val attrs =
	    case fval of
		NONE => []
	      | SOME v => [CONSTVAL (compileConstVal v)]
    in
	{flags=flags, name=name, ty=tyToJty(ty), attrs=attrs}
    end


(* ---------------- Add default initialiser to class if necessary  ---------------- *)

fun mkDefInit super = MDEF([M_ACCpublic],
		      NONE,
		      "<init>",
		      [],
		      MBODY ([VOIDdec(INVOKESPECIALop("this",MDESC(NONE,super^".<init>",[]),[]))],
			     [],
			     PRIMres VOIDres))

fun isInit (MDEF(_,_,name,_,_)) = name = "<init>"

fun addInit (cdef as CDEF(flags, C, super, interfaces, fdefs, mdefs, layout)) =
    if member C_ACCinterface flags then cdef  (* FIX: what if C is abstract?? *)
    else if List.exists isInit mdefs then cdef
    else
	let
	    val supername =
		case super of
		    NONE => "java.lang.Object"
		  | SOME s => s
	in
	    CDEF(flags, C, super, interfaces, fdefs, (mkDefInit supername)::mdefs, layout)
	end


fun fixACCsuper flags =
    if member C_ACCinterface flags then flags
    else if !setACCsuper
    then if member C_ACCsuper flags then flags
	 else C_ACCsuper::flags
    else List.filter (fn f => f <> Classdecl.C_ACCsuper) flags

    (* The JVM spec states that all compilers should produce classes with the
       ACCsuper bit set, so by default we add the ACCsuper flag.  In the unlikely
       event that a class file without the ACCsuper bit is required, set
       setACCsuper to false (this can be done with the gdf -unsuper option).
       Note that interfaces should NOT have ACCsuper set (although I can't
       see why it would matter). *)


(* -------------------------------- Main function -------------------------------- *)

infix 5 ::?
fun x::?l = case x of NONE => l | SOME y => y::l

fun compile filename cdef outputdir (printCert,dataLayout,tagOffset,thySyntax) richTypes =
let
    (* outputdir = name of directory in which to write class files;  must end with "/" *)

    val cdef' = addInit cdef

    (* We have to add the default initialiser  before typechecking so that calls
       to <init> resolve to current class (not superclass) *)

    val CDEF(flags, C, super, interfaces, fdefs, mdefs, layout) = cdef'

    val flags = fixACCsuper flags

    val () = setCurrentClass C flags

    val () = Typecheck.typeCheckClassDef cdef'

   (* We have to typecheck methods & fields immediately before
      compilation so that the type info is still available *)

    val fields = map compile_fdef fdefs
    val methods = map compile_mdef mdefs

    val javaLangObject = SOME (class{pkgs=["java","lang"], name="Object"})

    val superclass =
	case super of
	    NONE => javaLangObject
	  | SOME S => SOME (qnameToJClass S)



    val cl: class_decl =
	{ major = Classdecl.major, minor = Classdecl.minor,
	  flags = flags,
	  this = class{pkgs=[], name=C},
	  super = superclass,
	  ifcs = map qnameToJClass interfaces,
	  fdecls = fields,
	  mdecls = methods,
	  attrs=(Metadata.makeLayoutInfo layout)::?
		[SRCFILE(GrailUtils.basename filename),
		 Metadata.makeTypeInfo()]
	}

    val cp = Constpool.create();

    val os = BinIO.openOut (GrailUtils.makeFullFilename outputdir C  "class")
    fun emitword w = BinIO.output1 (os, w);
    val () = Classfile.emit emitword cp cl
    val () = BinIO.closeOut os


    (* optionally, also generate a .thy file out of the grail code *)
    val generateThy = not (thySyntax=0)

    val () = if generateThy then
               let
	           val fname = GrailUtils.makeFullFilename outputdir C "thy"
                   val thy = TextIO.openOut fname
	           val () = ToyGrailAbsyn.printGrailPROG thy cdef' (printCert,dataLayout,tagOffset,thySyntax)
	           val () = TextIO.closeOut thy
	       in
		   printToStdErr ("Wrote "^fname^"\n")
	       end
	     else ()

   val () = if printCert
              then CertGenP.makeCert1 C mdefs outputdir richTypes
				      (printCert,dataLayout,tagOffset,thySyntax)
              else () (* TextIO.print "Omitting certgen, as requested\n" *)
in
    C
end
