(* Grail abstract syntax *)

val required = () (* for mosmldep *)

type Var = string
type layout = string * (string * int * (string * string) list) list
(* Info about layout of Camelot type in Diamond class:
   (typename, [(constructor name, tag, [(field name, field type)])]) *)

datatype Ty =
	 BOOLEANty (* Usually int *)
       | BYTEty    (* 8-bit signed int *)
       | SHORTty   (* 16-bit signed int *)
       | INTty     (* 32-bit signed int *)
       | LONGty    (* 64-bit signed int *)
       | CHARty    (* 16-bit unsigned Unicode character *)
       | FLOATty   (* float *)
       | DOUBLEty  (* extended float *)
       | REFty of string
       | ARRAYty of Ty

type RTy = Ty option         (* Return type: NONE -> void *)

datatype Value =
	 VARval    of Var
       | BYTEval   of int
       | SHORTval  of int
       | INTval    of int  (* FIX: make this Int32.int: mosml 31-bit int might be too short *)
       | CHARval   of int
       | LONGval   of Int64.int
       | FLOATval  of Real32.real    (* These real types make rest of syntax fail to admit equality *)
       | DOUBLEval of Real64.real
       | STRINGval of string
       | CLASSval  of string  (* For Jclassconst *)
       | NULLval   of string * string option
	(* First component is object type,
	   2nd component is Camelot type, for theorem-proving *)

datatype UnaryOp = NEGop

datatype BinOp = ADDop | SUBop | MULop | DIVop | MODop
	       | ANDop | ORop  | XORop | SHLop | SHRop | USHRop  (* bitwise logic *)
	       | CMPop | CMPGop | CMPLop  (* Comparisons for floats &c *)

datatype Cast = dtof | dtoi | dtol | ftod | ftoi | ftol
	      | itob | itoc | itod | itof | itol | itos
	      | ltod | ltof | ltoi

datatype MethDesc = MDESC of RTy * string * Ty list
datatype FieldDesc = FDESC of Ty * string

datatype PrimOp =
    VALop of Value                                     (* Put a value on the stack *)
  | UNARYop of UnaryOp * Value                         (* Unary operator: so far just NEGop *)
  | IINCop of Var * int                                (* iinc: maybe we don't need this *)
  | BINop of BinOp * Value * Value                     (* Arithmetic/logical operation: see above *)
  | NEWop of MethDesc * Value list                     (* Create a new object & call an initialiser *)
  | CHECKCASTop of Ty * Var                            (* checkcast *)
  | INSTANCEop of Ty * Var                             (* instance check *)
  | INVOKESTATICop of MethDesc * Value list            (* Static method invocation *)
  | INVOKEDYNAMICop of Var * MethDesc * Value list     (* Dynamic method invocation *)
  | INVOKEINTERFACEop of Var * MethDesc * Value list   (* Interface method invocation *)
  | INVOKESPECIALop of Var * MethDesc * Value list     (* Superclass/initialiser invocation *)
  | INVOKEVIRTUALop of Var * MethDesc * Value list     (* Instance method invocation *)
  | GETFIELDop of Var * FieldDesc		       (* Get instance field *)
  | PUTFIELDop of Var * FieldDesc * Value              (* Set instance field *)
  | GETSTATICop of FieldDesc			       (* Get static field *)
  | PUTSTATICop of FieldDesc * Value		       (* Set static field *)
  | GETop of Value * Value			       (* Get array element *)
  | SETop of Value * Value * Value		       (* Set array element *)
  | LENGTHop of Var				       (* Length of array *)
  | NEWARRAYop of Ty * Value list                      (* Create empty array *)
  | CASTop of Cast * Value                             (* Casts between primitive types *)
  | LOCKop of Value                                    (* = monitorenter *)
  | UNLOCKop of Value                                  (* = monitorexit *)
  | THROWop of Var (* or Value? *)                     (* Throw an exception *)

datatype LetDec =
	 VALdec of Var * PrimOp
       | VOIDdec of PrimOp

datatype Test = EQ0 | NE0 | LT0 | LE0 | GT0 | GE0 | EQNULL | NENULL
datatype Comparison = EQcmp | NEcmp | LTcmp | LEcmp | GTcmp | GEcmp

datatype PrimRes =
	 VOIDres
       | OPres of PrimOp
       | FUNres of string * Var list

datatype Result =
	 PRIMres of PrimRes
       | CHOICEres of Value * Comparison * Value * PrimRes * PrimRes
       | TESTres of Test * Value * PrimRes * PrimRes
       | CASEres of Var * int * int * (int * string * Var list) list * (string * Var list)
       | LOOKUPres of Var * (int * string * Var list) list * (string * Var list)

datatype FunBody =
	 FUNbody of LetDec list * Result

datatype FunDec = FDEC of string * (Ty*Var) list * FunBody

datatype MethodBody = MBODY of LetDec list * FunDec list * Result
		    | NOBODY (* For interface methods *)

datatype MethodDef =
	 MDEF of Classdecl.method_access_flag list
		 * RTy                (* return type *)
		 * string             (* method name *)
		 * (Ty*Var) list      (* arguments with types *)
		 * MethodBody

datatype FieldDef = FDEF of Classdecl.field_access_flag list * Ty * string * Value option
(* Value option is initial value for interface fields *)

datatype ClassDef =
	 CDEF of Classdecl.class_access_flag list * string *
		 string option * string list * FieldDef list * MethodDef list
		 * layout list option
;

exception grailError of string;


(* ---------------- String representations of types ---------------- *)

fun tyToString t =
    case t of
	BOOLEANty => "boolean"
      | BYTEty    => "byte"
      | SHORTty   => "short"
      | INTty     => "int"
      | LONGty    => "long"
      | CHARty    => "char"
      | FLOATty   => "float"
      | DOUBLEty  => "double"
      | ARRAYty t => tyToString(t) ^ "[]"
      | REFty s   => s

fun rtyToString (NONE) = "void"
  | rtyToString (SOME ty) = tyToString ty


(* ---------------- Conversions from Grail representations to sml-jvm ---------------- *)

fun qnameToJClass name =
    if name ="" then Jvmtype.class{pkgs = [], name = ""} else
    let
	val l = String.tokens (fn c => (c = #".")) name
	val n = List.length l
	val pkgs = List.take (l,n-1)
	val name' = List.nth (l,n-1)
    in
	Jvmtype.class {pkgs = pkgs, name = name'}
    end

fun tyToJty t =
    case t of
	BOOLEANty  => Jvmtype.Tboolean
      | INTty      => Jvmtype.Tint
      | BYTEty     => Jvmtype.Tbyte
      | SHORTty    => Jvmtype.Tshort
      | CHARty     => Jvmtype.Tchar
      | LONGty     => Jvmtype.Tlong
      | FLOATty    => Jvmtype.Tfloat
      | DOUBLEty   => Jvmtype.Tdouble
      | REFty x    => Jvmtype.Tclass (qnameToJClass x)
      | ARRAYty ty => Jvmtype.Tarray (tyToJty ty)

and rtyToJty (SOME t) = SOME (tyToJty t)
  | rtyToJty NONE = NONE

fun tyToClassRef t =
    case t of REFty s => Bytecode.CLASS(qnameToJClass s)
	    | ARRAYty t => Bytecode.ARRAY (tyToJty t)
	    | _ => raise grailError ("tyToClassRef: unexpected type " ^ tyToString t)


(* -------------------------------- Functions on types and operators -------------------------------- *)

fun typeOfCast c =
    case c of
	dtof => FLOATty
      | dtoi => INTty
      | dtol => LONGty
      | ftod => DOUBLEty
      | ftoi => INTty
      | ftol => LONGty
      | itob => BYTEty
      | itoc => CHARty
      | itod => DOUBLEty
      | itof => FLOATty
      | itol => LONGty
      | itos => SHORTty
      | ltod => DOUBLEty
      | ltof => FLOATty
      | ltoi => INTty

fun typeOfCastArg c =
    case c of
	dtof => DOUBLEty
      | dtoi => DOUBLEty
      | dtol => DOUBLEty
      | ftod => FLOATty
      | ftoi => FLOATty
      | ftol => FLOATty
      | ltod => LONGty
      | ltof => LONGty
      | ltoi => LONGty
      | _ => INTty

fun numeric t =
    case t of
	BYTEty   => true
      | SHORTty  => true
      | INTty    => true
      | LONGty   => true
      | FLOATty  => true
      | DOUBLEty => true
      |  _ => false

fun isIntegral t =
    case t of
	BOOLEANty => true
      | BYTEty => true
      | SHORTty => true
      | INTty => true
      | _ => false

fun isReference t =
    case t of
	REFty _ => true
      | ARRAYty _ => true
      | _ => false

fun isArray t =
    case t of
	ARRAYty _ => true
      | _ => false

fun intOrLong t =
    case t of
	LONGty => true
      | _ => isIntegral t

fun logical oper =
    case oper of
	ANDop => true
      | ORop  => true
      | XORop => true
      | SHLop => true
      | SHRop => true
      | USHRop => true
      | _ =>  false

fun isShift oper =
    case oper of
	SHLop => true
      | SHRop => true
      | USHRop => true
      | _ => false

fun isComparison oper =
    case oper of
	CMPop => true
      | CMPGop => true
      | CMPLop => true
      | _ => false


(* ---------------- Fiddling with the boolean type ---------------- *)


(* The following stuff is to fix the bool/int problem.  The JVM does have
   a boolean type (Z in descriptors),  but it really just treats it as
   int.  If you get a boolean field then you can immediately add 77 to it
   without the verifier complaining;  booleans obtained as method return
   values are treated similarly. [There may be some special treatment for
   boolean arrays.]  The stuff below converts booleans to ints.  Whenever
   we declare (perhaps implicitly) a value of type boolean it gets treated
   as being of type int.  However,  boolean types are preserved in descriptors.
   I think this does the right thing;  it makes the Grail type system rather
   odd,  but it's consistent with the behaviour allowed by the JVM.  *)

(* FIX:  now I'm not so sure;  if a method declares a value of type boolean,
   what should its type be in local function declarations:  int or boolean? *)

(* Can the JVM even distinguish between local variables of type int and boolean? YES. *)

fun stackType t =
    case t of
	BOOLEANty => INTty (* Used when we create a var of type boolean *)
      | CHARty => INTty
      | SHORTty => INTty
      | _ => t

(* FIX: OTHERS *)

fun tysEqual x y = stackType x = stackType y

fun rtysEqual (SOME x) (SOME y) = tysEqual x y
  | rtysEqual NONE NONE = true
  | rtysEqual _ _ = false

(* Careful: this makes eg boolean = int.  This is OK;  these types only
   ever (?) have to be distinguished when comparing method and field
   descriptors. *)



(* ---------------- Variable and field descriptors for Isabelle output ---------------- *)

fun tyPrefix t =
    case t of
        BOOLEANty => "z"
      | BYTEty =>    "b"
      | CHARty =>    "c"
      | DOUBLEty =>  "d"
      | FLOATty =>   "f"
      | INTty =>     "v"
      | LONGty =>    "l"
      | SHORTty =>   "s"
      | REFty _ =>   "r"
      | ARRAYty _ => "a"

(* Make these consistent with JVM type descriptors? *)

fun uc1char s = String.str (Char.toUpper (String.sub(s,0)))

fun ucTyPrefix t = uc1char (tyPrefix t)


(************ Printing abstract syntax ************)
fun member x [] = false | member x (h::t) = x=h orelse member x t

val indent = ref "";
fun incIndent() = indent := (!indent) ^ "   ";
fun decIndent() = indent := String.extract(!indent, 3, NONE);


val ostream = ref TextIO.stdOut : TextIO.outstream ref
fun setOut x = ostream := x;
fun Pr x = TextIO.output(!ostream,x)
fun Prl s = TextIO.output(!ostream,s^"\n")
fun tab() = TextIO.output(!ostream, !indent)
fun nl(s) =  TextIO.output(!ostream, "\n" ^ (!indent) ^ s)

(* HWL was here *)
val reallyprint = ref true;
fun PrF os x = if (!reallyprint) then TextIO.output(os, x^" ") else ();
fun PrlF os x = if (!reallyprint) then TextIO.output(os, x^"\n") else ();
fun nlF os s =  if (!reallyprint) then TextIO.output(os, "\n" ^ (!indent) ^ s) else ();

fun castToString c =
    case c of
	dtof => "dtof"
      | dtoi => "dtoi"
      | dtol => "dtol"
      | ftod => "ftod"
      | ftoi => "ftoi"
      | ftol => "ftol"
      | itob => "itob"
      | itoc => "itoc"
      | itod => "itod"
      | itof => "itof"
      | itol => "itol"
      | itos => "itos"
      | ltod => "ltod"
      | ltof => "ltof"
      | ltoi => "ltoi"

fun cmpToString t =
    case t of
	EQcmp => "="
      | NEcmp => "<>"
      | LTcmp => "<"
      | LEcmp => "<="
      | GTcmp => ">"
      | GEcmp => ">="

fun testToString t =
    case t of
	EQ0 => "ifeq"
      | NE0 => "ifne"
      | LT0  => "iflt"
      | LE0 => "ifle"
      | GT0  => "ifgt"
      | GE0 => "ifge"
      | EQNULL => "ifnull"
      | NENULL => "ifnonnull"

fun unaryOpToString oper =
    case oper of
	NEGop => "neg"

fun binOpToString t =
    case t of
	ADDop  => "add"
      | SUBop  => "sub"
      | MULop  => "mul "
      | DIVop  => "div"
      | MODop  => "mod"
      | CMPop  => "cmp"
      | CMPGop => "cmpg"
      | CMPLop => "cmpl"
      | ANDop  => "and"
      | ORop   => "or"
      | XORop  => "xor"
      | SHLop  => "shl"
      | SHRop  => "shr"
      | USHRop => "ushr"

fun strOfInt n =
    let
	val s = Int.toString (Int.abs n)
    in
	if (n<0) then "-" ^ s else s
    end

val strOfInt32 = Int32.toString

val prInt = Pr o strOfInt

fun last n s = String.extract (s, size s - n, NONE)

fun valToString w =
    case w of
	VARval v => v
      | BYTEval b  => strOfInt b ^ "b"
      | SHORTval s => strOfInt s ^ "s"
      | INTval n   => Int32.toString (Int32.fromInt n)
      | LONGval l  => Int64.toString l ^ "L"
      | CHARval c  => (
	String.str (Char.chr c) handle Chr => raise grailError "character too big")
      | FLOATval f => "0fx" ^ Real32.toHexString f 
		      ^ " (* " ^ Real32.toString f ^ "f *)"
      | DOUBLEval d => "0dx" ^ Real64.toHexString d 
		       ^ " (* " ^ Real64.toString d ^ " *)"
      | STRINGval s => "\"" ^ s ^ "\""
       (* CAUTION:  string values can contain escape sequences.  We require that
          escape sequences are converted into literals during code emission.
          Disassembly should perform the reverse translation. *)
      | CLASSval c => "class " ^ c
      | NULLval(s,t) =>
	let in
	    case t of NONE => "null[" ^ s ^ "]"
		    | SOME d => "null[" ^ s ^ "]{" ^ d ^ "}"
	end


fun listToString toString separator l =
case l of [] => ""
        | [h] => toString h
        | h::t => (toString h) ^ separator ^ (listToString toString separator t)

fun id x = x

fun vecToString toString l = (* eg [1,2,5] -> "(1, 2, 5)" *)
    "("
    ^ (listToString toString ", " l)
    ^ ")"

val valVecToString = vecToString valToString
val tyVecToString =  vecToString tyToString

val strListToString = listToString (fn x => x)

fun prVec toString l =
       Pr (vecToString toString l)

val prVarVec = prVec id

fun valArrayToString a = "[" ^ listToString valToString "," a ^ "]"

val prTyVec = prVec tyToString

val prTyVarVec = prVec (fn (t,v) => (tyToString t ^ " " ^ v))

val prValVec = prVec valToString

fun nth([],_) = raise grailError "nth: list to short"
  | nth (h::_,0) = h
  | nth (_::t,n) = nth(t,n-1)



fun prCmp t = Pr (" " ^ cmpToString t ^ " ")
fun prTest t  = Pr (testToString t ^ " ")

fun prUnaryOp t = Pr (unaryOpToString t ^ " ")
fun prBinOp t   = Pr (binOpToString t ^ " ")

fun prValue v = Pr (valToString v)

fun prTy t = Pr (tyToString t)

fun prRTy r = Pr (rtyToString r)

fun cflagToString f =   (* Maybe we don't want to print out all of these *)
    case f of
	Classdecl.C_ACCpublic     => "public"
      | Classdecl.C_ACCprotected  => "protected"
      | Classdecl.C_ACCprivate    => "private"
      | Classdecl.C_ACCstatic     => "static"
      | Classdecl.C_ACCfinal      => "final"
      | Classdecl.C_ACCsuper      => ""    (* Handled specially *)
      | Classdecl.C_ACCinterface  => ""    (* Handled specially *)
      | Classdecl.C_ACCabstract   => "abstract"
      | Classdecl.C_ACCsynthetic  => "synthetic"
      | Classdecl.C_ACCannotation => "annotation"
      | Classdecl.C_ACCenum       => "enum"

fun prCflag f =
    case f of
	Classdecl.C_ACCpublic     => Pr "public "
      | Classdecl.C_ACCprotected  => Pr "protected "
      | Classdecl.C_ACCprivate    => Pr "private "
      | Classdecl.C_ACCstatic     => Pr "static "
      | Classdecl.C_ACCfinal      => Pr "final "
      | Classdecl.C_ACCsuper      => ()    (* Handled specially *)
      | Classdecl.C_ACCinterface  => ()    (* Handled specially *)
      | Classdecl.C_ACCabstract   => Pr "abstract "
      | Classdecl.C_ACCsynthetic  => Pr "synthetic "
      | Classdecl.C_ACCannotation => Pr "annotation "
      | Classdecl.C_ACCenum       => Pr "enum "


fun fflagToString f =
    case f of
	Classdecl.F_ACCpublic    => "public"
      | Classdecl.F_ACCprivate   => "private"
      | Classdecl.F_ACCprotected => "protected"
      | Classdecl.F_ACCstatic    => "static"
      | Classdecl.F_ACCfinal     => "final"
      | Classdecl.F_ACCvolatile  => "volatile"
      | Classdecl.F_ACCtransient => "transient"
      | Classdecl.F_ACCsynthetic => "synthetic"
      | Classdecl.F_ACCenum      => "enum"

fun prFflag f = Pr (fflagToString f ^ " ")

fun mflagToString f =
    case f of
	Classdecl.M_ACCpublic       => "public"
      | Classdecl.M_ACCprivate      => "private"
      | Classdecl.M_ACCprotected    => "protected"
      | Classdecl.M_ACCstatic       => "static"
      | Classdecl.M_ACCfinal        => "final"
      | Classdecl.M_ACCsynchronized => "synchronised"
      | Classdecl.M_ACCbridge       => "bridge"
      | Classdecl.M_ACCvarargs      => "varargs"
      | Classdecl.M_ACCnative       => "native"
      | Classdecl.M_ACCabstract     => "abstract"
      | Classdecl.M_ACCstrictfp     => "strictfp"
      | Classdecl.M_ACCsynthetic    => "synthetic"

fun prMflag f = Pr (mflagToString f ^ " ")


fun methDescToString (MDESC(rty, name, types)) =
    "<" ^ rtyToString rty
    ^ " " ^ name
    ^ tyVecToString types ^ "> "

val prMethDesc = Pr o methDescToString

fun fieldDescToString (FDESC(ty, name)) =
    "<" ^ tyToString ty ^ " "
    ^ name ^ ">"

val prFieldDesc = Pr o fieldDescToString

fun newToString (MDESC(rty, name, types), vals) =
    "new <" ^ name
    ^ " " ^ tyVecToString types
    ^ "> " ^ valVecToString vals

val prNew = Pr o newToString

fun prPrimOp p =
    case p of
	VALop v => prValue v
      | UNARYop(u, v) => (prUnaryOp u; prValue v)
      | IINCop (v,n) => Pr ("iinc " ^ v ^ " " ^ Int.toString n)
      | BINop(b, v, w) => (prBinOp b; prValue v; Pr " "; prValue w)
      | NEWop(m,v) => prNew (m, v)
      | CHECKCASTop(t,v) => Prl ("checkcast " ^ tyToString t ^ " " ^ v )
      | INSTANCEop(t,v) =>  Prl ("instanceof " ^ tyToString t ^ " " ^ v )
      | INVOKESTATICop(mdesc, args) =>
	( Pr "invokestatic "; prMethDesc mdesc; prValVec args )
      | INVOKEDYNAMICop(v, mdesc, args) =>
	( Pr ("invokedynamic " ^ v ^ " "); prMethDesc mdesc; prValVec args )
      | INVOKEINTERFACEop(v, mdesc, args) =>
	( Pr ("invokeinterface " ^ v ^ " "); prMethDesc mdesc; prValVec args )
      | INVOKESPECIALop(v, mdesc, args) =>
	( Pr ("invokespecial " ^ v ^ " "); prMethDesc mdesc; prValVec args )
      | INVOKEVIRTUALop(v, mdesc, args) =>
	( Pr ("invokevirtual " ^ v ^ " "); prMethDesc mdesc; prValVec args )
      | GETFIELDop(v, fdesc) =>
	( Pr ("getfield " ^ v ^ " " ); prFieldDesc fdesc )
      | PUTFIELDop(v, fdesc, w) =>
	( Pr ("putfield " ^ v ^ " "); prFieldDesc fdesc; Pr " "; prValue w )
      | GETSTATICop(fdesc) =>
	( Pr "getstatic "; prFieldDesc fdesc )
      | PUTSTATICop(fdesc,w) =>
	( Pr "putstatic "; prFieldDesc fdesc; Pr " "; prValue w)
      | SETop(aa, i, a) =>
	( Pr "set "; prValue aa; Pr " "; prValue i; Pr " "; prValue a)
      | LENGTHop aa =>
	( Pr "length "; print aa)
      | NEWARRAYop(ty,l) =>
	let
	    val a = valArrayToString l
	    val dim = length l
	    fun fixty t =
		if dim=1 then t
		else
		    let fun f t n =
			    if n=0 then t
			    else case t of
				     ARRAYty t' => f t' (n-1)
				   | _ => raise grailError ("invalid newarray: " ^ tyToString ty ^ "/" ^ a)
		    in
			f t dim
		    end
	in
	    (Pr "new "; prTy (fixty ty); Pr " ";Pr a)
	end
      | GETop(a,i) =>
	( Pr "get "; prValue a; Pr " "; prValue i)
      | CASTop(c,v) =>
	( Pr (castToString c); Pr " "; prValue v)
      | LOCKop v => (Pr "lock "; prValue v)
      | UNLOCKop v => (Pr "unlock "; prValue v)
      | THROWop v => Pr ("throw " ^ v)

fun primOpToString p =
    case p of
	VALop v => valToString v
      | UNARYop(u, v) => unaryOpToString u ^ " " ^ valToString v
      | IINCop (v,n) => "iinc " ^ v ^ " " ^ Int.toString n
      | BINop(b, v, w) => binOpToString b ^ " " ^ valToString v ^ " " ^ valToString w
      | NEWop (md,vs) => newToString (md,vs)
      | CHECKCASTop(t,v) => "checkcast " ^ tyToString t ^ " " ^ v
      | INSTANCEop(t,v) =>  "instanceof " ^ tyToString t ^ " " ^ v
      | INVOKESTATICop(mdesc, args) =>
	"invokestatic " ^ methDescToString mdesc ^ valVecToString args
      | INVOKEDYNAMICop(v, mdesc, args) =>
	"invokedynamic " ^ v ^ " " ^ methDescToString mdesc ^ valVecToString args
      | INVOKEINTERFACEop(v, mdesc, args) =>
	"invokeinterface " ^ v ^ " " ^ methDescToString mdesc ^ valVecToString args
      | INVOKESPECIALop(v, mdesc, args) =>
	"invokespecial " ^ v ^ " " ^ methDescToString mdesc ^ valVecToString args
      | INVOKEVIRTUALop(v, mdesc, args) =>
	"invokevirtual " ^ v ^ " " ^ methDescToString mdesc ^ valVecToString args
      | GETFIELDop(v, fdesc) =>
	"getfield " ^ v ^ " " ^ fieldDescToString fdesc
      | PUTFIELDop(v, fdesc, w) =>
	"putfield " ^ v ^ " " ^ fieldDescToString fdesc ^ " " ^ valToString w
      | GETSTATICop(fdesc) =>
	"getstatic " ^ fieldDescToString fdesc
      | PUTSTATICop(fdesc,w) =>
	"putstatic " ^ fieldDescToString fdesc ^ " "^ valToString w
      | SETop(aa, i, a) =>
	"set " ^ valToString aa ^ " " ^ valToString  i ^ " " ^ valToString a
      | LENGTHop aa =>
	"length " ^ aa
      | NEWARRAYop(ty,l) =>
	let
	    val a = valArrayToString l
	    val dim = length l
	    fun fixty t =
		if dim=1 then t
		else
		    let fun f t n =
			    if n=0 then t
			    else case t of
				     ARRAYty t' => f t' (n-1)
				   | _ => raise grailError ("invalid newarray: " ^ tyToString ty ^ "/" ^ a)
		    in
			f t dim
		    end
	in
	    "new " ^ tyToString (fixty ty) ^ " " ^ a
	end
      | GETop(a,i) =>
	"get " ^ valToString a ^ " " ^ valToString i
      | CASTop(c,v) =>
	castToString c ^ " " ^ valToString v
      | LOCKop v => "lock " ^ valToString v
      | UNLOCKop v => "unlock " ^ valToString v
      | THROWop v => "throw " ^ v

val prPrimOp = Pr o primOpToString

fun prPrimRes r =
    case r of
	OPres p => prPrimOp p
      | VOIDres => Pr "()"
      | FUNres (name, args) => (Pr name; prVarVec args)


fun prCase s (n, name, args) =
    (
     nl s;
     prInt n;
     Pr " => ";
     Pr name;
     prVarVec args
)

fun prDefault (name, args) =
    (
     nl "| ";
     Pr "_ => ";
     Pr name;
     prVarVec args
)

fun prResult r =
    case r of
	PRIMres p => (incIndent(); nl ""; prPrimRes p; decIndent())
      | CHOICEres (v, cmp, w, p1, p2) =>
	(
	 incIndent();
	 nl "if "; prValue v; prCmp cmp; prValue w;
	 nl "then "; prPrimRes p1;
	 nl "else "; prPrimRes p2;
	 decIndent()
	)
      | TESTres (tst, v, p1, p2) =>
	(
	 incIndent();
	 nl "";
	 prTest tst; prValue v;
	 nl "then "; prPrimRes p1;
	 nl "else "; prPrimRes p2;
	 decIndent()
	)
      | CASEres (v, low, high, cases, default) =>
	(
	 incIndent();
	 nl "case "; Pr v; Pr " in [";
	 prInt low; Pr ".."; prInt high;
	 Pr "] of";
	 incIndent ();
	 case cases of
	     [] => ()
	   | h::t => (prCase "  " h;  app (prCase "| ") t);
	 prDefault default;
	 decIndent ();
	 decIndent()
	)
      | LOOKUPres (v, cases, default) =>
	(
	 incIndent();
	 nl "case "; Pr v; Pr " of";
	 incIndent ();
	 case cases of
	     [] => ()
	   | h::t => (prCase "  " h;  app (prCase "| ") t);
	 prDefault default;
	 decIndent ();
	 decIndent()
	)

fun letDecToString d =
    case d of
	VALdec(id, p) => "val " ^ id ^ " = " ^ primOpToString p
      | VOIDdec p => "val () = " ^ primOpToString p

fun prLetDec (VALdec(id, p)) =
    ( nl("val " ^ id ^ " = "); prPrimOp p )
  | prLetDec (VOIDdec p) =
    ( nl "val () = "; prPrimOp p)

fun prFunBody (FUNbody(letdecs, result)) =
    case letdecs of
	[] => prResult result
      | _ =>
	(nl "let";
	 incIndent();
	 app prLetDec letdecs;
	 decIndent();
	 nl "in";
	 prResult result;
	 nl "end"
	 )

fun prFunDec (FDEC (fname, params, fbody)) =
    (
     Pr "\n";
     nl("fun " ^ fname );
     prTyVarVec params;
     Pr " =";
     prFunBody fbody
)

fun prMethodBody b =
    case b of
	MBODY(letDecs, funDecs, result) =>
	(
	 Pr " =";
	 nl "let";
	 incIndent();
	 app prLetDec letDecs;
	 app prFunDec funDecs;
	 decIndent();
	 nl "in";
	 prResult result;
	 nl "end\n"
	)
      | NOBODY => Pr "\n"

fun prMethodDef (MDEF(flags, rty, name, params, mbody)) =
    (
     nl "method ";
     app prMflag flags;
     prRTy rty;
     Pr (" " ^ name ^ " ");
     prTyVarVec params;
     prMethodBody mbody
     )


fun prFieldDef (FDEF(flags, ty, name, ival )) =
    (
     nl "field ";
     app prFflag flags;
     prTy ty;
     Pr " ";
     case ival of
	 NONE => Prl name
       | SOME v => (Pr name; Pr " = "; prValue v; nl "")
    )

fun prIntfs _ [] = ()
  | prIntfs isIf ifs =
    (
     if isIf then Pr " extends " else Pr " implements ";
     Pr (listToString id " " ifs)
    )

fun prSuper (SOME cl) = Pr (" extends " ^ cl)
  | prSuper NONE = ()

fun prClassDef (CDEF (flags, cname, super, intfs, fdefs, mdefs, layout)) =
    (if
	 member Classdecl.C_ACCinterface flags
     then
	 (  (* interface *)
	  app prCflag flags;
	  Pr "interface ";
	  Pr cname;
	  prIntfs true intfs
	 )
     else
	 (   (* class *)
	  if not (member Classdecl.C_ACCsuper flags)
	  then Pr "(* ACCsuper flag NOT set *)\n"
	  else ();
	  app prCflag flags;
	  Pr "class ";
	  Pr cname;
	  prSuper super;
	  prIntfs false intfs
	 );

     Pr " {";
     incIndent();
     app prFieldDef fdefs;
     app prMethodDef mdefs;
     decIndent();
     Prl "}";
     Option.app prLayout layout
    )

and prLayout l =
    let
	fun listToString toString separator l =
	    case l of [] => ""
		    | [h] => toString h
		    | h::t => (toString h) ^ separator ^ (listToString toString separator t)

	fun prConLayout (C, tag, fields) =
	    (
	     tab();
	     Pr C;
	     Pr ": $tag = ";
	     Pr (Int.toString tag);
	     Pr ", $fields = [";
	     Pr (listToString (fn (s,t) => "<" ^ s ^ ", " ^ t ^ ">") ", " fields);
	     Prl "]"
	    )

	fun pr1type (tname, l) =
	    (incIndent();
	     tab();
	     Pr "type ";
	     Prl tname;
	     incIndent ();
	     app prConLayout l;
	     decIndent ();
	     Pr "\n";
	     decIndent ()
	    )
    in
	(
	 Prl "\nlayout {";
	 app pr1type l;
	 Prl "}\n"
	)
    end
;




(* Matthew's CoreGrail Abstract Syntax printing functions *)

val names = ref [];

(* da: really should record correct types here as well --- might as well accumulate Isabelle
   syntax, in fact *)
fun addName (n:string) = if (not (List.exists (fn x => x=n) (!names))) andalso
			    (not (n = "()")) andalso   (* da: avoid generating bad Isabelle syntax *)
			    (not (n = "<init>"))
			 then names := n::(!names) else ();

fun printNames [] = ()
  | printNames (h::t) = (Pr (h^" :: \"string\"\n"); printNames t)

fun printNamesF os [] = ()
  | printNamesF os (h::t) = (PrF os (h^" :: \"string\"\n"); printNamesF os t)


fun printCoreGrailPROG (MDEF (_, _, _, _, body)) =
(
  Prl "theory ACoreGrailProgram = CoreGrailAbsyn:\n";
  Pr "constdefs prog :: \"Prog\"\n\"prog ==";
  incIndent();
  nl "PROG (";
  incIndent(); nl "";
  printCoreGrailMBODY body;
  decIndent();
  nl ")";
  decIndent();
  Prl "\"";

  Prl "";
  Pr ("(* Move the next "^Int.toString((length (!names))+1)^" to before the \"constdefs prog...\" line *)\n");
  Pr "consts\n";
  printNames(!names);
  Prl "\nend"
)

and printCoreGrailMBODY (MBODY (letDecs, funDecs, result)) =
(
  Pr "MBODY";
  incIndent();

  (if null letDecs then nl "[]" else (
  nl "["; incIndent();
  printCoreGrailLetDecs letDecs;
  decIndent(); Pr "]"));

  (if null funDecs then nl "[]" else (
  nl "["; incIndent();
  printCoreGrailFunDecs funDecs;
  decIndent(); Pr "]"));

  nl ""; printCoreGrailResult result;

  decIndent()
)
  | printCoreGrailMBODY NOBODY = Pr "(* INTERFACE METHOD DECLARATION *)"  (* FIX *)

and printCoreGrailLetDecs [] = ()
  | printCoreGrailLetDecs (h::[]) = printCoreGrailLetDec h
  | printCoreGrailLetDecs (h::t) = (printCoreGrailLetDec h; Pr ","; printCoreGrailLetDecs t)

and printCoreGrailLetDec (VALdec (name, primop)) =
( addName(name);
  nl ("(VALdec "^name^" ");
  printCoreGrailPrimOp primop;
  Pr ")"
)
  | printCoreGrailLetDec (VOIDdec (primop)) =
(
  Pr "(VOIDdec ";
  printCoreGrailPrimOp primop;
  Pr ")"
)

and printCoreGrailFunDecs [] = ()
  | printCoreGrailFunDecs (h::[]) = printCoreGrailFunDec h
  | printCoreGrailFunDecs (h::t) = (printCoreGrailFunDec h; Pr ","; printCoreGrailFunDecs t)

and printCoreGrailFunDec (FDEC (name, args, body)) =
( addName(name);
  nl ("(FDEC "^name);
  incIndent();

  (if args=[] then nl "[]" else (
  nl "["; incIndent();
  printCoreGrailFunArgs args;
  decIndent(); Pr "]"));

  printCoreGrailFunBody body;

  decIndent(); Pr ")"
)

and printCoreGrailFunArgs [] = ()
  | printCoreGrailFunArgs ((ty,v)::[]) = Pr ("("^(if (ty=INTty) then "INTty" else "(NOT SUPPORTED)")^","^v^")")
  | printCoreGrailFunArgs ((ty,v)::t) = (Pr ("("^(if (ty=INTty) then "INTty" else "(NOT SUPPORTED)")^","^v^")"); Pr ","; printCoreGrailFunArgs t)

and printCoreGrailFunBody (FUNbody (letDecs, result)) =
(
  nl "(FUNbody";
  incIndent();

  (if null letDecs then nl "[]" else (
  nl "["; incIndent();
  printCoreGrailLetDecs letDecs;
  decIndent(); nl "]"));

  (*nl "("; incIndent();*)
  nl "";
  printCoreGrailResult result;
  (*decIndent(); nl ")";*)
  decIndent(); Pr ")"
)

and printCoreGrailResult (PRIMres primres) =
(
  Pr "(PRIMres"; printCoreGrailPrimRes primres; Pr ")"
)
  | printCoreGrailResult (CHOICEres (v, test, v', primres, primres')) =
(
  Pr "(CHOICEres";
  incIndent();
  nl "(CONDhead ";
  printCoreGrailValue v; Pr " ";
  printCoreGrailTest test; Pr " ";
  printCoreGrailValue v';
  nl ")";
  nl "";
  printCoreGrailPrimRes primres;
  nl "";
  printCoreGrailPrimRes primres';
  decIndent();
  nl ")"
)
  | printCoreGrailResult (TESTres (test, name, p1, p2)) =
    (
     Pr "TESTres not implemented in CoreGrail\n"
    )
  | printCoreGrailResult (CASEres (name, low, high, cases, default)) =
    (
     Pr "CASEres not implemented in CoreGrail\n"
    )
  | printCoreGrailResult (LOOKUPres (name, cases, default)) =
    (
     Pr "LOOKUPres not implemented in CoreGrail\n"
    )

and printCoreGrailPrimOp (VALop v) = (Pr "(VALop "; printCoreGrailValue v; Pr ")")
  | printCoreGrailPrimOp (BINop (b,v,v')) =
(
  Pr "(BINop ";
  printCoreGrailBinOp b; Pr " ";
  printCoreGrailValue v; Pr " ";
  printCoreGrailValue v';
  Pr ")"
)
  | printCoreGrailPrimOp (INVOKESTATICop (M,cs)) = Pr "INVOKESTATICop..."
  | printCoreGrailPrimOp _ = Pr "(NOT SUPPORTED)"

and printCoreGrailPrimRes (OPres p) = (Pr "(OPres "; printCoreGrailPrimOp p; Pr ")")
  | printCoreGrailPrimRes VOIDres = Pr "VOIDres"
  | printCoreGrailPrimRes (FUNres (s,args)) = (Pr ("(FUNres "^s^" ["); printCoreGrailArgNames args; Pr "])")

and printCoreGrailArgNames [] = ()
  | printCoreGrailArgNames (h::[]) = Pr h
  | printCoreGrailArgNames (h::t) = (Pr (h^","); printCoreGrailArgNames t)

and printCoreGrailValue (VARval v) = (addName v; Pr ("(VARval "^v^")"))
  | printCoreGrailValue (INTval i) = Pr ("(INTval "^Int.toString(i)^")")
  | printCoreGrailValue (STRINGval s) = Pr ("(STRINGval "^s^")")
  | printCoreGrailValue (NULLval (s,t)) =
    (
     case t of NONE => Pr ("(NULLval "^s^")")
	     | SOME d => Pr ("(NULLval "^s^"){"^d^"}")
    )
  | printCoreGrailValue _ = Pr "(NOT SUPPORTED)"

and printCoreGrailTest EQcmp = Pr "EQUALStest"
  | printCoreGrailTest LTcmp = Pr "LESStest"
  | printCoreGrailTest _ = Pr "(NOT SUPPORTED)"

and printCoreGrailBinOp ADDop = Pr "ADDop"
  | printCoreGrailBinOp SUBop = Pr "SUBop"
  | printCoreGrailBinOp MULop = Pr "MULop"
  | printCoreGrailBinOp _ = Pr "(NOT SUPPORTED)"
;

