{
 (* Lexical analyser for Grail.  Adapted from example in Moscow ML distribution. *)

open Lexing Parser;

exception LexicalError of string * int * int (* (message, loc1, loc2) *)

fun lexerError lexbuf s =
     raise LexicalError (s, getLexemeStart lexbuf, getLexemeEnd lexbuf);

val commentStart = ref 0;  (* Start of outermost comment being scanned *)

fun commentNotClosed lexbuf =
     raise LexicalError ("Comment not terminated",
                         !commentStart, getLexemeEnd lexbuf);

val commentDepth = ref 0;  (* Current comment nesting *)


fun getString lexbuf =
     let
	val s = getLexeme lexbuf
     in case String.fromString s of
         SOME _ => quoteval s (* return the unconverted string:  we're only checking for errors *)
       | NONE => lexerError lexbuf "Badly-formed string literal"
     end

exception noSuchWord

val keyword_table: (string,token) Polyhash.hash_table
       = Polyhash.mkTable(Polyhash.hash, op=) (64, noSuchWord)

(* Hashtable for identifiers *)

val () =
   List.app (fn (str,tok) => Polyhash.insert keyword_table (str, tok))
 [
  (* Arithmetic/logical *)

  ("neg",  NEG),
  ("add",  ADD),
  ("mul",  MUL),
  ("sub",  SUB),
  ("div",  DIV),
  ("mod",  MOD),
  ("iinc", IINC),
  ("and",  AND),
  ("or",   OR),
  ("xor",  XOR),
  ("shr",  SHR),
  ("shl",  SHL),
  ("ushr", USHR),

  (* Casts *)

  ("dtof", DTOF),
  ("dtoi", DTOI),
  ("dtol", DTOL),
  ("ftod", FTOD),
  ("ftoi", FTOI),
  ("ftol", FTOL),
  ("itob", ITOB),
  ("itoc", ITOC),
  ("itod", ITOD),
  ("itof", ITOF),
  ("itol", ITOL),
  ("itos", ITOS),
  ("ltod", LTOD),
  ("ltof", LTOF),
  ("ltoi", LTOI),

  (* Methods *)

  ("invokestatic",    INVOKESTATIC),
  ("invokevirtual",   INVOKEVIRTUAL),
  ("invokespecial",   INVOKESPECIAL),
  ("invokeinterface", INVOKEINTERFACE),
  ("invokedynamic",   INVOKEDYNAMIC),

  (* Fields *)

  ("getfield",  GETFIELD),
  ("putfield",  PUTFIELD),
  ("getstatic", GETSTATIC),
  ("putstatic", PUTSTATIC),

  (* Objects *)

  ("new",        NEW),
  ("instanceof", INSTANCEOF),
  ("checkcast",  CHECKCAST),

  ("lock",   LOCK),
  ("unlock", UNLOCK),
  ("throw",  THROW),

  (* Arrays *)

  ("makearray", MAKE),
  ("get",       GET),
  ("set",       SET),
  ("length",    LENGTH),
  ("empty",     EMPTY),

  (* Branches *)

  ("cmp",      CMP),
  ("cmpg",     CMPG),
  ("cmpl",     CMPL),
  ("ifeq",     IFEQ),
  ("ifne",     IFNE),
  ("iflt",     IFLT),
  ("ifle",     IFLE),
  ("ifgt",     IFGT),
  ("ifge",     IFGE),
  ("ifnull",   IFN),
  ("ifnonnull", IFNN),

  (* Constants *)

  ("null",  NULL),
  ("true",  TRUE),
  ("false", FALSE),

  (* Types *)

  ("byte",    BYTE),
  ("short",   SHORT),
  ("int",     INT),
  ("long",    LONG),
  ("char",    CHAR),
  ("float",   FLOAT),
  ("double",  DOUBLE),
  ("boolean", BOOLEAN),
  ("ref",     REF),
  ("string",  STRING),
  ("void",    VOID),

  (* Keywords *)

  ("alias",  ALIAS),
  ("class",  CLASS),
  ("field",  FIELD),
  ("method", METHOD),
  ("let",    LET),
  ("fun",    FUN),
  ("in",     IN),
  ("and",    AND),
  ("end",    END),
  ("val",    VAL),
  ("if",     IF),
  ("then",   THEN),
  ("else",   ELSE),
  ("case",   CASE),
  ("of",     OF),

  ("extends",    EXTENDS),
  ("implements", IMPLEMENTS),

  (* Flags *)

  ("abstract",     ABSTRACT),
  ("annotation",   ANNOTATION),
  ("bridge",       BRIDGE),
  ("enum",         ENUM),
  ("final",        FINAL),
  ("interface",    INTERFACE),
  ("native",       NATIVE),
  ("private",      PRIVATE),
  ("protected",    PROTECTED),
  ("public",       PUBLIC),
  ("static",       STATIC),
  ("strictfp",     STRICTFP),
  (*  ("super",        SUPER),*)   (* Now set by default: see Compile.setACCsuper *)
  ("synchronized", SYNCHRONIZED),
  ("synchronised", SYNCHRONIZED),
  ("synthetic",    SYNTHETIC1),    (* SYNTHETIC seems to be reserved *)
  ("transient",    TRANSIENT),
  ("varargs",      VARARGS),
  ("volatile",     VOLATILE),

  (* Miscellaneous *)

  ("type", TYPE),
  ("layout", LAYOUT),
  ("$tag", TAG),
  ("$fields", FIELDS)

]

fun mkKeyword lexbuf =
  let val s = getLexeme lexbuf in
    Polyhash.find keyword_table s
    handle noSuchWord => stringval s
  end

fun chop n s = String.extract (s, n, NONE)                (* Throw away first n characters *)
fun chopr n s = String.extract (s, 0, SOME (size s - n))  (* Throw away last n characters *)

fun getInt lexbuf =  (* FIX this when we put Int32.int in INTconst *)
    let val s = getLexeme lexbuf
    in case Int.fromString s 
	     handle Overflow => raise Fail ("Constant too big: " ^ s)
	 of 
            NONE   => lexerError lexbuf "internal error"
	  | SOME i => i
    end 
}


let hexstring = ([`0`-`9`]|[`a`-`f`]|[`A`-`F`])+
let intstring = `-`?[`0`-`9`]+
let decstring = `-`?[`0`-`9`]+`.`[`0`-`9`]+([`e``E`]`-`?[`0`-`9`]+)?

rule Token = parse
    [` ` `\t` `\n` `\r`]     { Token lexbuf }

  | intstring `b`   { int8 (getInt lexbuf) }

  | intstring `s`   { int16 (getInt lexbuf) }

  | intstring       { int32 (getInt lexbuf) }

  | intstring [`l``L`]   { int64(Int64.fromString (chopr 1 (getLexeme lexbuf))) }

  | decstring[`f``F`]
	     { case Real.fromString (getLexeme lexbuf) of
	       NONE   => lexerError lexbuf "internal error"
	       | SOME i => float32 (Real32.fromReal i)
	       }

  | decstring
	     { case Real.fromString (getLexeme lexbuf) of
	       NONE   => lexerError lexbuf "internal error"
	       | SOME i => float64 (Real64.fromReal i)
	       }

  | decstring[`d``D`]
	     { case Real.fromString (getLexeme lexbuf) of
	       NONE   => lexerError lexbuf "internal error"
	       | SOME i => float64 (Real64.fromReal i)
	       }

  | "0bx" hexstring
	     { int8hex (chop 3 (getLexeme lexbuf)) }

  | "0sx" hexstring
	     { int16hex (chop 3 (getLexeme lexbuf)) }

  | "0x" hexstring
	     { int32hex (chop 2 (getLexeme lexbuf)) }

  | "0xi" hexstring
	     { int32hex (chop 3 (getLexeme lexbuf)) }

  | "0lx" hexstring
	     { int64hex  (chop 3 (getLexeme lexbuf)) }

  | "0fx" hexstring
	     { float32hex (chop 3 (getLexeme lexbuf)) }

  | "0dx" hexstring
	     { float64hex (chop 3 (getLexeme lexbuf)) }

  | [`a`-`z``A`-`Z``$``?`][`a`-`z``A`-`Z``0`-`9``_``.``#``$``'`]*(`<`[`a`-`z`]+`>`)?`'`*
                        { mkKeyword lexbuf }
  | `<`[`a`-`z`]+`>` { stringval (getLexeme lexbuf) }
  | `"`([^`"`]|"\\\"")*`"`
                        { getString lexbuf }  (* I'm not entirely confident about this *)
  | "(*"                { commentStart := getLexemeStart lexbuf;
                          commentDepth := 1;
                          SkipComment lexbuf; Token lexbuf }
  | "//"                { DiscardRestOfLine lexbuf; Token lexbuf }
  | `=`                 { EQUALS }
  | "<>"                { NOTEQUAL }
  | `<`                 { LESS }
  | "<="                { LESSEQ }
  | `>`                 { GREATER }
  | ">="                { GREATEREQ }
  | "=>"                { ARROW }
  | ".."                { DOTDOT }
  | `(`                 { LPAREN }
  | `)`                 { RPAREN }
  | `{`                 { LBRACE }
  | `}`                 { RBRACE }
  | `[`                 { LSQ }
  | `]`                 { RSQ }
  | `|`                 { BAR }
  | `,`                 { COMMA }
  | `:`                 { COLON }
  | `_`                 { UNDERSCORE }
  | eof                 { EOF }
  | _                   { lexerError lexbuf "Illegal symbol in input" }

and SkipComment = parse
     "*)"               { commentDepth := !commentDepth - 1;
                          if !commentDepth = 0 then ()
                          else SkipComment lexbuf
                        }
   | "(*"               { commentDepth := !commentDepth + 1;
                          SkipComment lexbuf }
   | (eof | `\^Z`)      { commentNotClosed lexbuf }
   | _                  { SkipComment lexbuf }

and DiscardRestOfLine = parse
     (eof | `\n`) { () }
   | _ { DiscardRestOfLine lexbuf }

;
