(* Various tables storing persistent information for typechecking
   and compilation. *)

open Classdecl

open GrailAbsyn
open GrailUtils

(* -------------------- Errors with locations -------------------- *)

(* We can't use textual locations in the source file since we often have
   abstract code generated by Camelot *)

fun getPackage C =
    let
	val l = String.tokens (fn c => (c = #".")) C
	val n = List.length l
    in
	if n=1 then []
	else List.take(l,n-1)
    end

val currentClass = ref ""
val currentPackage = ref []: string list ref
val currentClassFlags = ref []: Classdecl.class_access_flag list ref

fun setCurrentClass s l =
    (currentClass := s;
     currentPackage := getPackage s;
     currentClassFlags := l)

fun currentClassHas f = member f (!currentClassFlags)
fun inCurrentPackage C = getPackage C = !currentPackage

val currentMethod = ref ""
fun setCurrentMethod s = currentMethod := s

val currentField = ref ""
fun setCurrentField s = currentField := s

val currentFunction = ref ""
fun setCurrentFunction s = currentFunction := s

fun here () =
    let val f = !currentFunction
    in
	"! [class " ^ !currentClass
	^ ", method " ^ !currentMethod
	^ (if f = "" then "]:"
	  else ", function " ^ f ^ "]:")
    end

fun localError s =
    let val () = errPrint (here() ^ "\n! " ^ s ^ "\n")
    in quit ()
    end

fun primOpErr p s =
    let val () = errPrint (here() ^ "\n!  " ^ GrailAbsyn.primOpToString p ^ "\n! " ^ s ^ "\n")
    in quit ()
    end

fun letErr d s =
    let val () = errPrint (here() ^ "\n!  " ^ GrailAbsyn.letDecToString d ^ "\n! " ^ s ^ "\n")
    in quit ()
    end

fun methodErr s =
    let val () = errPrint ("[class " ^ !currentClass
			^ ", method " ^ !currentMethod
			^ "]\n! " ^ s ^ "\n")
    in quit ()
    end

fun fieldErr s =
    let val () = errPrint ("[class " ^ !currentClass
			^ ", field " ^ !currentField
			^ "]\n! " ^ s ^ "\n")
    in quit ()
    end

fun fieldWarning s =
    let val () = errPrint ("[class " ^ !currentClass
			^ ", field " ^ !currentField
			^ "]\n! Warning: " ^ s ^ "\n")
    in ()
    end

fun classErr s =
    let val () = errPrint ("[class " ^ !currentClass ^"]\n! " ^ s ^ "\n")
    in quit ()
    end

fun classWarning s =
    let val () = errPrint ("[class " ^ !currentClass ^"]\n! Warning: " ^ s ^ "\n")
    in ()
    end

(* ---------------------------- Local variables and labels ---------------------------- *)

fun lookup  v (ref l) =
    let fun
	lookup' l =
	case l of
	    [] => NONE
	  | (w,p)::tail => if v=w then SOME p else lookup' tail
in
    lookup' l
end

fun find v h = case lookup v h of
    NONE => localError ("Can't find variable/function " ^ v)
  | SOME p => p

(* Local Variables and labels for function declarations and calls *)


local

    type ident = string

    (* Local variables *)

    val localsref = ref Localvar.freshLocals : Localvar.locals ref

    fun newLocal1 () =
        let val (locals, loc) = Localvar.nextVar1 (!localsref)
        in localsref := locals; loc
	end

    fun newLocal2 () =
        let val (locals, loc) = Localvar.nextVar2 (!localsref)
        in localsref := locals; loc
	end

    val registers: (ident * (Ty * Localvar.index)) list ref
	= ref []  (* (name, (type, varpos)) *)

    (* Labels *)

    val nextlab = ref Label.freshLabels : Label.labels ref
    val (_,cs) = Label.newLabel(!nextlab)   (* Horrible *)
    val (_,ce) = Label.newLabel(!nextlab)   (* Horrible *)
    val funLabels: (ident * (RTy * (Ty*string) list * Label.label)) list ref
	= ref []

in
    val codeStart = ref cs
    val codeEnd = ref cs;  (* Both of these are initial values which will never be used *)

    fun resetLocals () = (localsref := Localvar.freshLocals; registers := [])
    fun numLocals() = Localvar.count (!localsref)  (* careful: some may be 64 bits wide *)

    fun lookupVar v = lookup v registers
    fun findVar v = find v registers

    fun saveVar v ty = 
	let
	    val loc =
		case ty of
		    LONGty => newLocal2()
		  | DOUBLEty => newLocal2()
		  | _ => newLocal1()
	in
	    registers := (v, (stackType ty, loc))::(!registers)
	end


    fun dumpLocals () =
	let
	    val l = !registers
	    fun dl [] n = ()
	      | dl ((name,_)::t) n =
		(BasicIO.print ((Int.toString n) ^ ": " ^ name ^"\n");
		 dl t (n-1))
	in dl l (List.length l - 1)
	end

    fun getLocals () =
	let
	    fun get' [] = []
	      | get'  ((id,(ty,loc))::t) =
		{from=(!codeStart), thru=(!codeEnd), name=id, ty=tyToJty(ty), index=loc}::get'(t)
	in
	    LOCALVAR(get' (!registers))
	end


    (* Labels *)

    fun newLabel () = (* We need this to be globally visible so that
		         we can generate new labels for "if ..." *)
        let val (labels, lab) = Label.newLabel (!nextlab)
        in nextlab := labels; lab
	end

    fun resetLabels () =
	(nextlab := Label.freshLabels;
	 codeStart := newLabel();
	 codeEnd := newLabel();
	 funLabels := [])

    fun lookupFun f = lookup f funLabels
    fun findFun f = find f funLabels
    fun getFunLabels () = rev (!funLabels)

    fun declareFuntype ty (FDEC (name, params,_)) =
	case lookupFun name of
	    SOME _ => methodErr ( "Trying to redeclare function " ^ name )
	  | NONE => let (* val () = vprint("Declaring function " ^ name ^ "\n") *)
			val loc = newLabel()
		    in
			funLabels := (name, (ty, params, loc))::(!funLabels)
		    end

end
