(* Typecheck.sml:  Grail types *)

val strict = ref false
fun setStrictTypeChecking b = strict := b

(* If strict is set to false then we don't typecheck things involving Java
   classes (in particular there is no subtype check for objects).  This is
   an order of magnitude faster, but we run the risk of producing non-verifiable
   bytecode. *)

val lax = ref false
(* If true then don't check scope for local function args;  useful
   when we've got variable consolidation turned off *)

open GrailAbsyn
open GrailUtils
open Classloader
open Persistent

val currentRtype = ref NONE: GrailAbsyn.RTy ref

(* Return type for current method; temporary(?) expedient to deal with
   type of THROWop *)

(* ----------------------------- Scope checking ------------------------------ *)
(* ---- We could possibly do this in the parser,  but that would be messy ---- *)

fun scopeCheck p vars = (* Check vars in primop are in scope *)
    let
	fun checkvar v =
	    if member v vars orelse (!lax) then ()
	    else primOpErr p ("Variable " ^ v ^ " not in scope")
	   (* There were problems here before we had minimal Grail argument
              lists: hopefully this is OK now. *)

	fun checkval w =
	    case w of (VARval v) => checkvar v
	  | _ => ()
    in
	case p of
	    VALop w                   => checkval w
	  | UNARYop (_,v)             => checkval v
	  | IINCop (v,_)              => checkvar v
	  | BINop(_,v,w)              => (checkval v; checkval w)
	  | NEWop(_,vs)               => app checkval vs
	  | CHECKCASTop (_,v)         => checkvar v
	  | INSTANCEop  (_,v)         => checkvar v
	  | INVOKESTATICop (_ ,l)     => app checkval l
	  | INVOKEVIRTUALop (v,_,l)   => (checkvar v; app checkval l)
	  | INVOKEDYNAMICop (v,_,l)   => (checkvar v; app checkval l)
	  | INVOKESPECIALop (v,_,l)   => (checkvar v; app checkval l)
	  | INVOKEINTERFACEop (v,_,l) => (checkvar v; app checkval l)
	  | GETFIELDop  (v,_)         => checkvar v
	  | PUTFIELDop (v,_,w)        => (checkvar v; checkval w)
	  | GETSTATICop _             => ()
	  | PUTSTATICop (_,v)         => checkval v
          | NEWARRAYop (t,l)          => ()
          | GETop (a,i)               => (checkval a; checkval i)
          | SETop (a,i,v)             => (checkval a; checkval i; checkval v)
          | LENGTHop a                => checkvar a
          | CASTop (c,v)              => checkval v
	  | LOCKop v                  => checkval v
	  | UNLOCKop v                => checkval v
	  | THROWop v                 => checkvar v
    end

fun scopeCheckLetDecs l vars =
    case l of
	[] => vars
      | h::t =>
	case h of
	    VOIDdec p => (scopeCheck p vars; scopeCheckLetDecs t vars)
	  | VALdec (v,p) => (scopeCheck p vars; scopeCheckLetDecs t (v::vars))

fun scopeCheckArg vars f v =
    if member v vars then ()
    else localError ("Function " ^ f ^ " applied to unknown variable " ^ v)

fun scopeCheckPrimres vars r =
    case r of
	VOIDres => ()
      | OPres p => scopeCheck p vars
      | FUNres (f, args) => app (scopeCheckArg vars f) args

fun scopeCheckCase vars (key, f, args) = app (scopeCheckArg vars f) args

fun scopeCheckResult vars result =
    if !lax then ()
    (* If we have the -n Camelot flag set then we get dodgy arglists;
       this lets them through *)
    else case result of
	     PRIMres r => scopeCheckPrimres vars r
	   | CHOICEres (_,_,_, r1, r2) => (scopeCheckPrimres vars r1;
					   scopeCheckPrimres vars r2)
	   | TESTres (_,_, r1, r2) => (scopeCheckPrimres vars r1;
				       scopeCheckPrimres vars r2)
	   | CASEres (_,_,_,cases, (df,dargs)) => (app (scopeCheckCase vars) cases;
						   app (scopeCheckArg vars df) dargs)
	   | LOOKUPres (_,cases, (df,dargs)) =>  (app (scopeCheckCase vars) cases;
						  app (scopeCheckArg vars df) dargs)


fun scopeCheckFunDec (FDEC (fname, params, FUNbody (lets, result))) =
    let
	val () = setCurrentFunction fname
	val args = map snd params
	val vars = scopeCheckLetDecs lets args
    in
	scopeCheckResult vars result
    end

fun scopeCheckMdef  (MDEF(flags, rettype, mname, params,mbody)) =
    case mbody of
	NOBODY => ()
      | MBODY (letdecs,fundecs,result) =>
	let
	    val () = setCurrentFunction ""
	    val argnames =
		if member M_ACCstatic flags
		then map snd params
		else "this":: map snd params
		     
	  val ivars = scopeCheckLetDecs letdecs argnames
	  val () = scopeCheckResult ivars result
	in
	    app scopeCheckFunDec fundecs
	end




(* ------------------------------- Subtyping --------------------------------- *)

fun isSubclass t1 t2 =
    if !strict
    then
	Classloader.isSubclass t1 t2
    else true

fun subtype t1 t2 = 
    case (t1,t2) of
	(REFty c1, REFty c2)  => isSubclass c1 c2
      | (ARRAYty u1, ARRAYty u2) => subtype u1 u2
      | (ARRAYty u1, REFty t) => t = "java.lang.Object" 
				 orelse t = "java.lang.Cloneable" 
				 orelse t = "java.io.Serializable"
      | _ => GrailAbsyn.tysEqual t1 t2
	     
(* What about array operations??? *)

fun typeMsg v vty ty =
    (case ty of
	 REFty _ => "Subtype of " ^ tyToString ty
       | _ => "Type " ^ tyToString ty)
    ^ " was expected, but "
    ^ valToString v
    ^ " has type " ^ tyToString vty


fun mToString C m tys =
    C ^ "." ^ m ^ "(" ^ listToString tyToString "," tys ^ ")"



(* ------------------------- Type inference  ------------------------- *)

val inCurrentPackage = Persistent.inCurrentPackage

val currentClassHas = Persistent.currentClassHas (* Does current class have flag f? *)

fun declareVar v ty =
    case Persistent.lookupVar v of
	SOME (t1,_) =>
	if subtype ty t1 then ()  (* FIX:  is this OK??? What about funcalls? *)
	else localError ( "Trying to reuse variable "
			  ^ v ^ ": " ^ tyToString t1
			  ^ " with new type " ^ tyToString ty)
      | NONE => Persistent.saveVar v ty
		
fun declareParams l = app (fn (t,v) => declareVar v t) l
		      
fun typeOfVar v =
    case lookupVar v of
	NONE => localError ("Can't infer type of variable " ^ v)
      | SOME (t,_) => t

fun typeOfValue w =
    case w of
	VARval v      => typeOfVar v
      | BYTEval _     => BYTEty
      | SHORTval _    => SHORTty
      | INTval _      => INTty
      | LONGval _     => LONGty
      | CHARval _     => CHARty
      | FLOATval _    => FLOATty
      | DOUBLEval _   => DOUBLEty
      | STRINGval _   => REFty "java.lang.String"
      | CLASSval _    => REFty "java.lang.Class"
      | NULLval (t,_) => REFty t

fun valAndType v = valToString v ^ ": "  ^ tyToString (typeOfValue v)
val varAndType  = valAndType  o VARval

fun valAndType' v = "(" ^ valAndType v ^ ")"
val varAndType' = valAndType' o VARval

fun typeOfPrimop p =
let
(*    val () = print (GrailAbsyn.primOpToString p ^ "\n")*)
    fun err s = primOpErr p s

    val assignable = subtype  
(*
    fun assignable S T = (* Assignment compatibility: 2.6.7 *)
	case S of 
	    REFty s => 
	    if isClass s then subclass S T
	    else T = Object orelse subclass S T
	  | ARRAYty SC =>
	    (
	     case T of
		 REFty t => 
		 if isClass t then t = "java.lang.Object"
		 else t = "java.lang.Cloneable" orelse t = "java.io.Serializable"
	       | ARRARty TC => assignable SC TC
	    )
	  | _ => FIX: narrowing
*)		      
    fun resolveClass C =
	let
	    val () = Classloader.resolveClass C
	in
	    if member C_ACCpublic (getCFlags C)
	       orelse inCurrentPackage C then ()
	    else err ("Class " ^ C ^ " is inaccessible from current class")
	end

    fun checkValTy (v, ty) =
	let val vty = typeOfValue v in
	    if subtype vty ty then ()
	    else err (typeMsg v vty ty)
	end

    fun checkArgTys args tys =
	if length args <> length tys
	then err "Argument length mismatch"
	else ListPair.app checkValTy (args,tys)

    fun splitQualName qname =
	let
	    val mname = List.last (String.fields (fn c => c = #".") qname)
	in
	    (truncate qname (1+ size mname), mname)
	end

    fun refty v =
	case typeOfVar v of
	    REFty c => c
	  | ARRAYty _ => "java.lang.Object"  (* FIX: eek *)
	  | _ => primOpErr p (v ^ " has non-reference type " ^ tyToString (typeOfVar v))

    fun abstractClass C = member C_ACCabstract (getCFlags C)

    fun fToString C f ty = C ^ "." ^ f ^ ": " ^ tyToString ty


    fun defaultFieldAccess flgs =
	let
	    fun is flg = member flg flgs
	in
	    is F_ACCprotected
	    orelse not (is F_ACCpublic
			orelse is F_ACCprotected
			orelse is F_ACCprivate)
	end

    fun resolveField C f ty =  (* FIX: must look in superinterfaces of C as well *)
	let
	    val () = resolveClass C

	    datatype result = NO of string list | YES of string * Member 


	    fun lookAtInterface I done =
		if member I done then NO done
		else 
		    case seekField f ty I of
			SOME f' => (vprint ("Found " ^ f ^ " in interface " ^ I ^ "\n"); YES (I,f'))
		      | NONE => lookAtInterfaces (getInterfaces I) (I::done)
	      

	    and lookAtInterfaces l done =
		case l of [] => NO done
			| I::t => 
			  if member I done then lookAtInterfaces t done
			  else case lookAtInterface I done of 
				   NO done' => lookAtInterfaces t done'
				 | y => y

	    and lookAtClass A done =
		case seekField f ty A of
		    SOME f' => (vprint ("Found " ^ f ^ " in class " ^ A ^ "\n"); (A,f'))
		  | NONE => case lookAtInterfaces (getInterfaces A) done of 
				YES r => r
			      | NO done => 
				case getSuperclass A of
				    SOME S => lookAtClass S done
				  | NONE => err ("Couldn't resolve field " ^ fToString C f ty)

	    val (C',f') = lookAtClass C []

	    val fflags = fieldFlags f'
	    fun fieldIs flg = member flg fflags

	    val () = if not (member C_ACCpublic (getCFlags C) orelse inCurrentPackage C)
		     then err ("Class " ^ C ^ " is not accessible from " ^ !currentClass)
		     else if fieldIs F_ACCpublic
			     orelse (fieldIs F_ACCprotected
				     andalso isSubclass (!currentClass) C')
			     orelse (defaultFieldAccess fflags andalso inCurrentPackage C')
			     orelse (fieldIs F_ACCprivate andalso C' = !currentClass)
		     then ()
		     else err "Field is inaccessible from current class"
	in
	    (C', f')
	end


    fun checkPutStatic fname ty arg =
	let
	    val () = if assignable (typeOfValue arg) ty then ()
		     else err ("Can't store value of type " ^ tyToString (typeOfValue arg)
			       ^ " in field of type " ^ tyToString ty)
	in
	    if !strict then
		let

		    val (C,f) = splitQualName fname
		    val (C',f') = resolveField C f ty

		    val fflags = fieldFlags f'
		in
		    if not (member F_ACCstatic fflags)
		    then err "Field is non-static"
		    else if member F_ACCfinal fflags andalso C' <> !currentClass
		    then err "Trying to assign to final field in external class"
		    else ()  
	    (* FIX: JVM spec now says that only <clinit< can write to final fields,
	       but JVM allows any method in the current class to do it. *)

            (* FIX: interface fields??? *)
		end
	    else
		()
	end


    fun checkPutField v fname ty arg =
	let
	    val () = if assignable (typeOfValue arg) ty then ()
		     else err ("Can't store value of type " ^ tyToString (typeOfValue arg)
			       ^ " in field of type " ^ tyToString ty)
  	 (* FIX: assignment compatible : 2.6.7 *)

	    val (C,f) = splitQualName fname
	    val O = refty v
	    val () = if isSubclass O C then ()
		     else err (v ^ " does not have field " ^ fToString C f ty)
	 (* FIX: presumably we need something like this,  but the spec doesn't say so *)

	in
	    if !strict then
		let
		    val (C',f') = resolveField C f ty
		    val fflags = fieldFlags f'

		    val () = if member F_ACCprotected fflags
				andalso isSubclass (!currentClass) C'
			     then if isSubclass O (!currentClass) then ()
				  else err (v ^ " cannot access field " ^ fToString C' f ty)
			     else ()
		    val () = if member F_ACCprotected fflags
			     then if isSubclass O (!currentClass) then ()
				  else err "Illegal access to protected field"
			     else ()  (* See note at end of 5.4.4 *)
		in
		    if member F_ACCstatic fflags
		    then err "Field is static"
		    else if member F_ACCfinal fflags andalso C' <> !currentClass
		    then err "Trying to assign to final field in external class"
		    else ()  
 	    (* FIX: JVM spec now says that only <init> can write to final fields,
	       but JVM allows any method in the current class to do it. *)

		end
	    else
		()
	end


    fun checkGetStatic fname ty =
	if !strict then
	    let
		val (C,f) = splitQualName fname

		val (C',f') = resolveField C f ty

		val fflags = fieldFlags f'
	    in
		if not (member F_ACCstatic fflags)
		then err "Field is non-static"
		else ()  (* FIX:  we also have to deal with interface fields *)
	    end
	else ()


    fun checkGetField v fname ty =
	let
	    val O = refty v
	    val (C,f) = splitQualName fname
	    val () = if isSubclass O C then ()
		     else err (v ^ " does not have field " ^ fToString C f ty)

            (* FIX: presumably we need something like this,  but the spec doesn't say so *)
	in
	    if !strict then
		let
		    val (C',f') = resolveField C f ty

		    val fflags = fieldFlags f'
		    val () = if member F_ACCprotected fflags
			     then if isSubclass O (!currentClass) then ()
				  else err "Illegal access to protected field"
			     else ()  (* See note at end of 5.4.4 *)
		    val () = if member F_ACCprotected fflags
				andalso isSubclass (!currentClass) C'
			     then if isSubclass O (!currentClass) then ()
				  else err (v ^ " cannot access field " ^ fToString C' f ty)
			     else ()
		in
		    if member F_ACCstatic fflags
		    then err "Field is static"
		    else ()
		end
	    else ()
	end


    fun defaultMethodAccess flgs =
	let
	    fun is flg = member flg flgs
	in
	    is M_ACCprotected
	    orelse not (is M_ACCpublic
			orelse is M_ACCprotected
			orelse is M_ACCprivate)
	end


    fun checkMethodAccessible C m =
	let 
	    val flags = methFlags m
	    fun methodIs f = member f flags
	in
	    if not (member C_ACCpublic (getCFlags C) orelse inCurrentPackage C)
	    then err ("Class " ^ C ^ " is not accessible from " ^ !currentClass)
	    else
		if methodIs M_ACCpublic
		   orelse (methodIs M_ACCprotected
			   andalso isSubclass (!currentClass) C)
		   orelse (defaultMethodAccess flags andalso inCurrentPackage C)
		   orelse (methodIs M_ACCprivate andalso C = !currentClass)
		then ()
		else err "Method is inaccessible from current class"
	end



    fun resolveMethodI m rty tys intfs =
	let
	    fun doInterface J l done =
		if member J done then doList l done
		else
		    if not (isInterface J)
		    then err ("interface expected, but class " ^ J ^ " found")
		    else
			case seekMethod m rty tys J of
			    SOME m' => (vprint ("Found " ^ m ^ " in interface " ^ J ^ "\n");  (J,m'))
			  | NONE =>
			    case getSuperclass J of
				SOME J' => doInterface J' l (J::done)
			      | NONE => doList l (J::done)

	    and doList l done =
		case l of
		    [] => err ("Unable to resolve method " ^ m)
		  | I::t => doInterface I t done
	in
	    doList intfs ["java.lang.Object"]
	end

    fun resolveInterfaceMethod m I rty tys =
	if m = "<init>" orelse m = "<clinit>"
	then err ("Illegal invocation of " ^ m)
	else
	let
	    val () = resolveClass I

	    fun f A =
		let
		    val () = if member C_ACCinterface (getCFlags A)
			     then ()
			     else err ("interface required,  but class " ^ A ^  " found")
		in
		    case seekMethod m rty tys A of
			SOME m' => (vprint ("Found " ^ m ^ " in interface " ^ A ^ "\n"); (A,m'))
		      | NONE =>
			case getSuperclass A of
			    SOME A' => f A'
			  | NONE =>
			    case seekMethod m rty tys "java.lang.Object" of
				SOME m' => (vprint ("Found " ^ m ^ " in class java.lang.Object\n");
					    ("java.lang.Object",m'))
			      | NONE => err ("Couldn't resolve interface method " ^ mToString I m tys)
		end

		val (I',m') = f I
		val () = checkMethodAccessible I' m'
	in
	    (I', m')
	end

    fun checkInterfaceCall p v name rty tys args =
	let
	    val err = primOpErr p
	    val () = checkArgTys args tys
	    val C = refty v
	    val (I,m) = splitQualName name

	in
	    if !strict then
		let
		    val () = resolveClass C  (* SHOULD THIS BE DONE ELSEWHERE? *)

		    val (I',m') = resolveInterfaceMethod m I rty tys

		    val () = if isInterface C then
				 if isSubclass C I then ()
				 else err ("Can't invoke " ^ mToString I m tys
					   ^ " on interface " ^ v ^ " of type " ^ C)
			     else
				 if implements C I
				 then ()
				 else err ("Class of " ^ v ^ " is " ^ C
					   ^ ", which does not implement interface " ^ I)

		  (* FIX:  Maybe I' is a superinterface of one of C's interfaces
                           Maybe I' is a superinterface of one of C's superclasses *)


		  (* This is probably incorrect, but it's not clear from the JVM spec
		     what static constraints there are on the type of v.  If the object
		     v of class C implements the method m:() -> () which is declared
		     in the interface I then both

			invokevirtual v <void C.m()>()

		     and

		        invokeinterface v <void I.m()>()

		     seem to be accepted by the verifier (as long as C is declared
		     to implement I).  *)

		in
		    if (member M_ACCstatic (methFlags m'))
		    then err (mToString C m tys ^ " resolves to static method")
		    else ()
		end
	    else ()
	end

    fun resolveMethod m C rty tys =
	let
	    val () = resolveClass C

	    fun f A =
		let
		    val () = if member C_ACCinterface (getCFlags A)
			     then err ("class required,  but interface " ^ A ^  " found")
			     else ()
		in
		    case seekMethod m rty tys A of
			SOME m' => (vprint ("Found " ^ m ^ " in class " ^ A ^ "\n"); (A,m'))
		      | NONE =>
			case getSuperclass A of
			    SOME A' => f A'
			  | NONE => err ("Couldn't resolve method " ^ mToString C m tys)
		end

		val (C',m') = f C

		val () = if member M_ACCabstract (methFlags m')
			    andalso (not (abstractClass C'))
			 then err (m ^ " resolves to abstract method")
			 else checkMethodAccessible C' m'
	in
	    (C', m')
	end

        (* The JVM spec says we should also try to resolve the method in 
           the superinterfaces of C if we fail to find it in a superclass, 
           but in this case that doesn't appear to matter since we later 
           have to check that m is implemented in some superclass of the
           class of the object involved, and we should have found this 
           before getting to the interfaces. *)
           

    fun checkVirtualCall p v name rty tys args =
	let
	    val err = primOpErr p
	    val () = checkArgTys args tys

	    val (C,m) = splitQualName name

	    val () = if m = "<init>" orelse m = "<clinit>"
		     then err ("Illegal invocation of " ^ m)
		     else ()

	    val O = refty v
	    val () = if isSubclass O C then ()
		     else err ("No applicable method: " ^ v ^ " has type " ^ O)
	in
	    if !strict then
		let
		    val (C',m') = resolveMethod m C rty tys
		    val flags = methFlags m'
		    (*
		        val () = if member M_ACCprotected flags
			     then if isSubclass O (!currentClass) then ()
				  else err ("Illegal access to protected method " ^ name)
			     else ()  (* See note at end of 5.4.4 *)
		     *)

		    val () = if member M_ACCprotected flags 
				andalso isSubclass (!currentClass) C 
			     then 
				 if isSubclass O (!currentClass) then ()
				 else err ("Illegal access to protected method " ^ name)
			     else ()  (* See note at end of 5.4.4 & comments on invokevirtual *)


		(* FIX: I don't understand the spec here.  Consider 
			  invokevirtual o <Object Object.clone()> ()
		   The method clearly resolves to Object.clone,  which is protected.
		   Also, Object is a superclass of the current class.  Now the spec 
		   says that if Object.clone isn't declared in the same runtime package
                   as the current class (so the current class isn't in java.lang) then
                   the class of o must be a subclass of the current class.  If o is of
                   type int[] then it's not a subclass of the current class, but the 
                   invocation actually succeeds.  On the other hand, if o is a BigInteger
                   then we get a verification error saying "Bad access to protected data".
	           Now section 2.15 of the spec says that arrays override the clone method,
                   but I don't see how that does us any good. 

                   Update: it now seems that the fact that javac generates Object.clone
                   (together with the fact that it's accepted by the VM) is a bug.  It 
		   seems that you should in fact get int[].clone .  This really messes 
                   things up for us.

		 *)
		in
		    if (member M_ACCstatic (methFlags m'))
		    then err (mToString C m tys ^ " resolves to static method")
		    else ()
		end
	    else ()
	end

    fun checkStaticCall p name rty args tys =
	let
	    val err = primOpErr p

	    val () = checkArgTys args tys

	    val (C,m) = splitQualName name
	in
	    if !strict then
		let
		    val (C',m') = resolveMethod m C rty tys
		in
		    if (member M_ACCstatic (methFlags m')) then ()
		    else err (name ^ " resolves to instance method")
		end
	    else ()
	end

    fun checkSpecialCall p v name rty tys args =
	let
	    val err = primOpErr p
	    val () = checkArgTys args tys

	    val (C,m) = splitQualName name
	    val O = refty v
	    val () = if isSubclass O C then ()
		     else err ("No applicable method: " ^ v ^ " has type " ^ O)
            (* Not mentioned in spec?  *)

	(* We might be calling a superclass initializer *)
	(*
	    val () = if m = "<init>" then
			 err "Call to <init> not allowed in Grail: use 'new' instead"
		     else ()
	 *)

	in
	    if !strict then
		let
		    val (C',m') = resolveMethod m C rty tys

		    val m'flags = methFlags m'

		    val () = if member M_ACCprotected m'flags
				andalso isSubclass (!currentClass) C'
				andalso not (inCurrentPackage C')
			     then
				 if isSubclass O (!currentClass) then ()
				 else err "Access error for protected method"
			     else ()

		    val () = if currentClassHas C_ACCsuper
				andalso isSubclass (!currentClass) C'
				andalso m <> "<init>"
			     then
				 case getSuperclass (!currentClass) of
				     NONE => ()
				   | SOME S =>
				     let
					 val _ = resolveMethod m S rty tys
				     in
					 ()
				     end
			     else ()

		in
		    (*
		    if m = "<init>" andalso C <> C'
		    then err "Illegal initialiser invocation"
		    else
		    *)

		    if member M_ACCstatic m'flags
		    then err (mToString C m tys ^ " resolves to static method")
		    else if member M_ACCabstract m'flags
		    then err "method is abstract"
		    else ()
		end
	    else ()
	end

    fun	checkNew p t name tys args =
	let
	    val err = primOpErr p
	    val () = checkArgTys args tys
	    val () = if Option.isSome t then
			 err "Malformed 'new' instruction: non-void return type"
		     else ()
	in
	    if !strict then
		let
		    val C = name (* Constructor descriptor is different from methdesc *)
		    val m = "<init>"

		    val (C',m') = resolveMethod m C NONE tys
		    val () = if C' = C then ()
			     else err ("Initialiser found in superclass " ^ C'
				       ^ " of " ^ C)

		    val cflags = getCFlags C
		    val mflags = methFlags m'
		in
		    if member C_ACCinterface cflags
		    then err ("Trying to instantiate interface " ^ C')
		    else if member C_ACCabstract cflags
		    then err ("Trying to instantiate abstract class " ^ C')
		    else if member M_ACCstatic mflags
		    then err "Initialiser is static"
		    else if member M_ACCabstract mflags
		    then err "Initialiser is abstract"
		    else ()
		end
	    else ()
	end


in

    case p of
	VALop v => SOME(typeOfValue v)

      | UNARYop (oper, v) =>
	let
	    val t = typeOfValue v
	in
	    case oper of
		NEGop => if numeric t then SOME t
			 else err ("Can't apply " ^ unaryOpToString oper
					   ^ " to " ^ tyToString t)
	end

      | IINCop (v,n) =>
	let
	    val t = typeOfVar v
	    val () = if n < ~128 orelse n > 127 then
			 err ("iinc: " ^ Int.toString n ^ " doesn't fit into byte")
		     else ()
	    val () = if t <> INTty then
			 err ("Can't apply iinc to " ^ varAndType v)
		     else ()
	in
	    NONE
	end

      | BINop(oper,v,w) =>
	let val t1 = typeOfValue v
	    val t2 = typeOfValue w
	    fun typeErr () = err ("Type error: "
					  ^ binOpToString oper
					  ^ " " ^ valAndType' v
					  ^ " " ^ valAndType' w
					 )
	in
	    if isShift oper
	    then
		if t2 <> INTty then err ("Attempting to shift value "
						 ^ valToString v
						 ^ " by "
						 ^ tyToString t2)
		else if not (intOrLong t1)
		then err ("Attempting to shift value "
				  ^ valToString v
				  ^ " of type "
				  ^ tyToString t1)
		else SOME t1
	    else
		if not (tysEqual t1 t2) then typeErr () (* FIX: check use of tysEqual *)
		else if logical oper andalso not (intOrLong t1) then typeErr ()
		else if numeric t1
		then if isComparison oper
		     then SOME INTty
		     else SOME t1
		else typeErr ()
    end

  | CHECKCASTop (t,v) =>
    let
	val vty = typeOfVar v
    in
	if isReference vty
	   then if isReference t
		then SOME t
		else err (tyToString t ^ ": expected reference type")
	else err (varAndType' v ^ " does not have reference type")
    end

  | INSTANCEop (t,v) =>
    let
	val vty = typeOfVar v
    in
	if isReference vty
	then if isReference t
	     then SOME INTty
	     else err (tyToString t ^ ": expected reference type")
	else err (varAndType' v ^ " does not have reference type")
    end

  | PUTFIELDop (v,FDESC(ty,name),w)                => (checkPutField v name ty w; NONE)
	     (* v must be obj, must have field of correct type *)
  | PUTSTATICop (FDESC(ty,name),w)                 => (checkPutStatic name ty w; NONE)
  | GETFIELDop  (v,FDESC (ty,name))                => (checkGetField v name ty; SOME ty)
  | GETSTATICop (FDESC (ty,name))                  => (checkGetStatic name ty; SOME ty)
  | INVOKEDYNAMICop (v,MDESC(rty,name,tys),args)   => (checkArgTys args tys; rty)
  | INVOKESPECIALop (v,MDESC(rty,name,tys),args)   => (checkSpecialCall   p v name rty tys args; rty)
  | INVOKESTATICop (MDESC(rty,name,tys),args)      => (checkStaticCall    p name rty args tys; rty)
  | INVOKEVIRTUALop (v,MDESC(rty,name,tys),args)   => (checkVirtualCall   p v name rty tys args; rty)
  | INVOKEINTERFACEop (v,MDESC(rty,name,tys),args) => (checkInterfaceCall p v name rty tys args; rty)
  | NEWop (MDESC(t,name,tys),args)                 => (checkNew p t name tys args; SOME(REFty name))

  | GETop (aa,n) =>
    if isIntegral (typeOfValue n)
       then
	case typeOfValue(aa) of
	    ARRAYty ty => SOME ty
	  | _ => err ("First argument of get must be of array type (not "
			       ^ tyToString (typeOfValue aa) ^ ")")
    else err ("index " ^ valAndType' n ^ " should be of type int")
		   (* FIX: or maybe short/boolean/... *)
  | SETop (aa,n,w)   =>
    if isIntegral (typeOfValue n)
    then
	case typeOfValue aa  of
	    ARRAYty ty  =>
	    let val ty' = typeOfValue w
	    in
		if subtype ty' ty then NONE  (* FIX: maybe we can't identify eg int and short here *)
		else err (typeMsg w ty' ty)
	    end
	  | _ => err ("First argument of get must be of array type (not "
			       ^ tyToString (typeOfValue aa) ^ ")")
    else err ("index " ^ valAndType' n ^ " should be of type int")

  | LENGTHop a =>
    let val t = typeOfVar a
    in if isArray t then SOME INTty
       else err (a ^ " has type " ^ tyToString t ^ ", but length requires an array")
    end

  | NEWARRAYop(ty,l) =>
    let fun f t l =
	    case l of [] => t
		    | (_::r) => f (ARRAYty t) r
    in
	SOME (f ty l)
    end

  | CASTop (c,v) =>
    let
	val cty = typeOfCastArg c
	val vty = typeOfValue v
    in
	if tysEqual cty vty (* FIX: tysEqual ?? *)
	then SOME (typeOfCast c)
	else err (typeMsg v vty cty)
    end
  | LOCKop v =>
    if isReference (typeOfValue v)
    then NONE
    else err ("Cannot lock " ^ valAndType' v ^ ": reference type required")
  | UNLOCKop v =>
    if isReference (typeOfValue v)
    then NONE
    else err ("Cannot unlock " ^ valAndType' v ^ ": reference type required")
  | THROWop v =>
    let
	val exty = REFty "java.lang.Throwable"
	val vty = typeOfVar v
    in
	if subtype vty exty then !currentRtype
	else err (typeMsg (VARval v) vty exty)
    end

end

fun typeCheckLetDecs l =
    case l of
	[] => ()
      | h::t =>
	case h of
	    VOIDdec(p) =>
	    (
	     case typeOfPrimop p of
		 NONE => typeCheckLetDecs t
	       | _ => letErr h ("Non-unit value assigned to ()")
	    )
	  | VALdec (v,p) =>
	    (
	     case typeOfPrimop p of
		 NONE => letErr h ("Attempting to assign unit value to variable " ^ v)
	       | SOME ty =>  (declareVar v ty; typeCheckLetDecs t)
	    )


fun typeCheckFunCall fname params =
    let
	val (_,formalParams,_) = findFun fname
	val formalNames = map snd formalParams
    in
	if params <> formalNames then
	    localError ("Function " ^ fname ^ " is called with arguments " ^
			vecToString id params ^
			", but requires arguments " ^ vecToString id formalNames)

(*	else if map typeOfValue params <> map fst formalParams then
	    raise gdfError ("INTERNAL ERROR: type mismatch in function call") *)
(* Doesn't work, but shouldn't happen *)
	else ()
    end

fun typeOfFunCall fname args =
    (
     typeCheckFunCall fname args;
     case (lookupFun fname) of
	 NONE => (localError ("Type error: can't find function " ^ fname))
       | SOME(t, n, p) => t
    )

fun typeOfPrimRes r =
    case r of
	OPres p => typeOfPrimop p
      | VOIDres => NONE
      | FUNres (fname, params) => typeOfFunCall fname params

fun typeOfResult r =
    case r of
	PRIMres p => typeOfPrimRes p
      | CHOICEres (v,cmp,w, p1,p2) =>
	let val t = typeOfValue v
	    val t' = typeOfValue w
	    val () = if tysEqual t t' then ()
		     else localError ("Test arguments have different types: "
				      ^ valAndType v
				      ^ " " ^ cmpToString cmp
				      ^ " " ^ valAndType w)

	    val () = case cmp of
			 EQcmp => ()
		       | NEcmp => ()
		       | _ => if not (numeric t)
			      then
				  localError ("Numeric type error: "
					      ^ valAndType v
					      ^ " " ^ cmpToString cmp
					      ^ " " ^ valAndType w
					     )
			      else ()
	    val t1 = typeOfPrimRes p1
	    val t2 = typeOfPrimRes p2

	in
	    if rtysEqual t1 t2 then t1
	    else localError ("Test branches have different types: "
			     ^ rtyToString t1
			     ^ " and "
			     ^ rtyToString t2)
	end

      | TESTres (tst, v, p1,p2) =>
	let val t = typeOfValue v
	    val () = if isIntegral t orelse isReference t then ()
		     else localError ("Illegal test " ^ testToString tst ^ " applied to value "
				      ^ valAndType v)
	    val t1 = typeOfPrimRes p1
	    val t2 = typeOfPrimRes p2
	in
	    if rtysEqual t1 t2 then t1
	    else localError ("Test branches have different types ("
			     ^ rtyToString t1 ^ " and " ^ rtyToString t2)
	end

      | CASEres (v, low, high, cases, (df, dargs)) =>
	let
	    fun checkcases l count =   (* Also check structure of case statement *)
		case l of [] => count
			| (n, f, args)::t =>
			  if n<>count then localError ("bad key ("
						   ^ Int.toString n
						   ^ ") in case statement")
			  else checkcases t (count+1)

	    val () =
		if high < low
		then localError "Invalid interval in case statement"
		else
		    if (checkcases cases low) <> high+1
		    then localError "Wrong number of cases in case statement"
		    else ()

	    val t = typeOfFunCall df dargs
	    fun ok (_,f,a) =
		let val t1 = typeOfFunCall f a
		in if t1 = t then ()
		   else localError "cases have incomaptible return types"
		end
		(* FIX: can this ever happen?  All function calls are tail-calls,
                   so they return the same type as the method. *)

	    val () = app ok cases
	in
	    t
	end

      | LOOKUPres (v, cases, (df, dargs)) =>
	let
	    fun checkcases l done =
		case l of
		    [] => ()
		  | (n, _,_)::t =>
		    if member n done
		    then localError ("Repeated key (" ^ Int.toString n ^ ") in case statement")
		    else checkcases t (n::done)
		    (* FIX: Does the verifier care about repeats? *)

	    val () = checkcases cases []

	    val t = typeOfFunCall df dargs
	    fun ok (_,f,a) =
		let val t1 = typeOfFunCall f a
		in if t1 = t then ()
		   else localError "cases have incomaptible return types"
		end
	    val () = app ok cases
	in
	    t
	end

(* Next function just contains some common code for functions and methods *)

fun typeCheckFun rtype fname params lets result =
    let
	val () = declareParams params
	val () = typeCheckLetDecs lets
	val rtype' = typeOfResult result
    in
	if rtysEqual rtype rtype' then ()
	else localError(fname
 			^ " should return " ^ rtyToString rtype
			^ ", but actually returns " ^ rtyToString rtype')
    end


fun typeCheckFunDec rtype (FDEC (fname, params, FUNbody (lets, result))) =
    let
	val () = setCurrentFunction fname
    in
	typeCheckFun rtype fname params lets result
    end


(* Checks on field flags *)

fun tooMany a b c l =
    if member a l
    then if member b l then SOME(a,b)
	 else if member c l then SOME(a,c)
	 else NONE
    else if member b l andalso member c l then SOME(b,c)
    else NONE


(* -------------------- Typechecking field declarations -------------------- *)

local

    fun validIFieldflag f =
    case f of
	F_ACCpublic => ()
      | F_ACCstatic => ()
      | F_ACCfinal  => ()
      | F_ACCsynthetic => ()
      | _ => fieldErr ("Modifier '" ^ fflagToString f ^ "' not allowed in interface field")

    fun err2 p q =
	fieldErr ("Field cannot be both " ^ fflagToString p
		  ^ " and " ^ fflagToString q)

    fun check2 flags p q =
	if member p flags andalso member q flags
	then err2 p q
	else ()

in

fun typeCheckFdef (FDEF(flags, ty, name, fval)) =
    let
	val () = vprint ("Typechecking field " ^ name ^ "\n")
	val () = setCurrentField name
    in
	if currentClassHas C_ACCinterface
	then
	    case fval of
		NONE => fieldErr "Interface fields must be initialised"
	      | SOME v =>
		let
		    val () = app validIFieldflag flags
		    val () = if member F_ACCpublic flags
				andalso member F_ACCstatic flags
				andalso member F_ACCfinal flags
			     then ()
			     else fieldErr "Interface fields must be public, static and final"
		    val vty = typeOfValue v
		in
		    if subtype vty ty then ()
		    else fieldErr ("Bad field value: " ^ typeMsg v vty ty)
		end

	else (* normal field *)
	case fval of
	    NONE =>
 	    let in case tooMany F_ACCpublic F_ACCprivate F_ACCprotected flags
		    of NONE => check2 flags F_ACCfinal F_ACCvolatile
		     | SOME(p,q) => err2 p q
	    end
	  | SOME _ =>
	    if member F_ACCstatic flags then ()
	    else fieldWarning ("ConstantValue attribute ignored for instance fields [JVMspec, 4.7.2].\n"
			   ^ "! Instance fields should be initialised in <init>")
    end
end

(* Checks on method flags *)

fun checkMflags flags =
    let
	val () = case tooMany M_ACCpublic M_ACCprivate M_ACCprotected flags
		  of NONE => ()
		   | SOME(p,q) => methodErr ("Method cannot be both " ^ mflagToString p
						^ " and " ^ mflagToString q)
	val () = if member M_ACCnative flags andalso member M_ACCstrictfp flags
		 then methodErr "Method cannot be both native and strictfp"
		 else ()

	val () = if member M_ACCstatic flags andalso member M_ACCabstract flags
		 then methodErr "Static method cannot be abstract"
		 else ()

	fun g f = if member f flags
		  then methodErr ("Abstract method cannot be " ^ mflagToString f)
		  else ()

	val () = if member M_ACCabstract flags
		 then app g [M_ACCprivate, M_ACCstatic, M_ACCfinal,
			     M_ACCnative, M_ACCstrictfp, M_ACCsynchronized]
		 else ()

	fun h f = if member f flags
		  then methodErr ("Interface method cannot be " ^ mflagToString f)
		  else ()

	val () = if currentClassHas C_ACCinterface
		 then
		     if member M_ACCpublic flags
			andalso member M_ACCabstract flags
		     then
			 app h [M_ACCprivate, M_ACCprotected, M_ACCstatic, M_ACCfinal,
				M_ACCsynchronized, M_ACCnative, M_ACCstrictfp]
		     else
			 methodErr "Interface method must be public and abstract"
		 else ()

		 
    in
	()
    end

(* -------------------- Typechecking method declarations -------------------- *)

    fun resolveMethodI m rty tys intfs =
	let
	    fun doInterface J l done =
		if member J done then doList l done
		else
		    if not (isInterface J)
		    then classErr ("interface expected, but class " ^ J ^ " found")
		    else
			case seekMethod m rty tys J of
			    SOME m' => (vprint ("Found " ^ m ^ " in interface " ^ J ^ "\n");  SOME (J,m'))
			  | NONE =>
			    case getSuperclass J of
				SOME J' => doInterface J' l (J::done)
			      | NONE => doList l (J::done)

	    and doList l done =
		case l of
		    [] => NONE
		  | I::t => doInterface I t done
	in
	    doList intfs ["java.lang.Object"]
	end


    (* FIX: surprisingly the VM doesn't object if you override 
      'final void f ()' with 'int f ()';  Java doesn't like this.
       Also, Java won't let you override a static method with a 
       non-static one (and vide-versa),  but JVM doesn't care. *)

    fun resolveMethod m rty tys B =
	let
	    fun f C =
		let
		    val () = if member C_ACCinterface (getCFlags C)
			     then classErr ("class required,  but interface " ^ C ^  " found")
			     else ()
		in
		    case seekMethod m rty tys C of
			SOME m' => (vprint ("Found " ^ m ^ " in class " ^ C ^ "\n"); SOME (C,m'))
		      | NONE =>
			case getSuperclass C of
			    SOME C' => f C'
			  | NONE => resolveMethodI m rty tys (getInterfaces B)
		end

	    val () = resolveClass B
	in
	    case getSuperclass B of
		NONE => NONE
	      | SOME S => f S  (* We're looking for methods that already exist *)
	end


fun typeCheckMdef (mdef as MDEF(flags, rtype, name, params, mbody)) =
    let
	val () = vprint ("Typechecking method " ^ rtyToString rtype ^ " " ^ name 
			 ^ "(" ^ listToString tyToString "," (map #1 params) ^ ")\n")
	val () = setCurrentMethod name
	val () = checkMflags flags
	val tys = map #1 params

	val () = if !strict
		 then case resolveMethod name rtype tys (!currentClass)
		       of NONE => ()
			| SOME (C,m) => (* overriding C.m *)
		          let  
			      val mflags = methFlags m
			  in
			      if member M_ACCfinal mflags
				 orelse member M_ACCprivate mflags
				 orelse member C_ACCfinal (getCFlags C) 
			      then methodErr ("Cannot override final method "
					      ^ mToString C name tys)
			      else ()
			  end
		 else ()
    in

    case mbody of
	NOBODY =>
	let
	    val () = if member M_ACCabstract flags then
			 if currentClassHas C_ACCabstract then ()
			 else methodErr "Defining class must be declared to be abstract"
		     else methodErr "Unimplemented method must be abstract"

	    val () = if member M_ACCstatic flags
		     then methodErr "Interface method cannot be static"
		     else ()
	in
	    ()
	end

      | MBODY (letdecs,fundecs,result) =>
	let
	    val () = resetLocals ()
	    val () = resetLabels ()

	    val () = app (declareFuntype rtype) fundecs
	    val () = scopeCheckMdef mdef

	    val params =
		if member M_ACCstatic flags then params
		else (REFty(!currentClass), "this")::params
		 (* instance method - fake declaration of "this" *)

	    val () = setCurrentMethod name
	    val () = setCurrentFunction ""

	    val () = checkMflags flags

	    val () = if name = "<init>"
		     then
			 let
			     fun k f =
				 if member f flags
				 then fieldErr ("Initiliaser cannot be " ^ mflagToString f)
				 else ()
			 in
			     app k [M_ACCstatic, M_ACCfinal, M_ACCsynchronized,
				    M_ACCbridge, M_ACCnative, M_ACCabstract]
			 end
		     else ()

	    val() = if member M_ACCnative flags
		    then methodErr "Native method cannot have implementation"
		    else ()

	    val () = if member M_ACCabstract flags
		     then methodErr "Abstract method cannot have implementation"
		     else if currentClassHas C_ACCinterface
		     then 
			 if name = "<init>" orelse name = "<clinit>" then ()
			 else methodErr "Method in interface class must be abstract"
		     else ()

           (* FIX: there's an odd comment in the appendix to the spec that says
	      that interfaces can have methods other than <init> and <clinit>.
              does it mean abstract methods? *)
           (* javac doesn't seem to allow you to define initialisers *)

	    val () = typeCheckFun rtype ("method " ^ name) params letdecs result
	    val () = app (typeCheckFunDec rtype) fundecs
	in
	    ()
	end
end

(* Check that interfaces are all valid and fully implemented *)

(* FIX: If the superclass S is abstract and implements some interface I
   then we have to check that any methods inherited from I via S are
   also implemented.  On the other hand,  it looks as if this isn't
   checked by the verifier.  You can (maybe) claim to implement an
   interface but as long as you don't actually call any methods
   from it you're OK. *)

fun checkInterfaces C =
    let
	val Cmethods = allMethods C

	fun sameMethod (METHOD(name, rty, args, flags)) (METHOD(name', rty', args', flags')) =
	    name' = name
	    andalso args' = args
	    andalso rty' = rty
	    andalso member M_ACCstatic flags' = member M_ACCstatic flags  (* See other sameMethod later *)
	  | sameMethod _ _ = classErr "Unexpected argument in Typcheck.sameMethod"


	fun ok I m = 
	    if List.exists (sameMethod m) Cmethods then ()
	    else classWarning ("Class " ^ C ^ " fails to implement interface method " ^ fullName I m)

	fun checkBranch I done =
	    if member I done then done
	    else
		let
		    val Imethods = allMethods I
		    val () = app (ok I) Imethods
		    val supers = getInterfaces I
		in
		    checkList supers (I::done)
		end

	and checkList l done =
	    case l of
		[] => []
	      | i::t =>
		let
		    val () = resolveClass i  (* also resolves superinterfaces *)
		    val () = if isInterface i then ()
			     else classErr ("Purported interface " ^ i ^ " is a class")
		    val done' = checkBranch i done
		in
		    checkList t done'
		end

	val _ =	checkList (getInterfaces C) []

	(* FIX: We also need to check that interfaces for superclass are implemented.
	   The problem is that a superclass S of C can implement an interface I
           but declare a method m from I to be abstract.  C then either has to
           implement m or declare it abstract.  
	       What happens if S implements m but then some class U with C < U < S
	   declares m abstract?  Is m then regarded as an interface method which
           C has to deal with?
	       This is probably specified for Java,  but it's not clear what the JVM 
           should do. Anyway,  the JVM doesn't seem to check that interfaces are
           actually implemented;  if you try to invoke an interface method I.m and 
           C doesn't say that it implements I then you get a verification error, but
           you can leave methods from I unimplemented and get away with it as 
           long as you don't try to invoke any of them. *)
    in
	()
    end


fun validInterfaceFlag f =
    case f of
	C_ACCinterface  => ()
      | C_ACCpublic     => ()
      | C_ACCabstract   => ()
      | C_ACCannotation => ()
      | _ => classErr ("Modifier '" ^ cflagToString f ^ "' not allowed in interface class")

fun checkCflags flags =
    if member C_ACCinterface flags
    then
	let val () = app validInterfaceFlag flags
	in
	    if currentClassHas C_ACCabstract then ()
	    else classErr "Interface class must be declared to be abstract"
            (* ??? spec says interface must be abstract, but then that it cannot be
               both final and abstract *)
	end
    else
	if member C_ACCannotation flags
	then classErr "annotation class must be interface"
	else ()
(* Should check for "public private" etc,  but the verifier
   doesn't seem to object to this *)



fun checkSuper C done =
    if member C done then classErr "Circularity error"
    else
	let
	    val super = getSuperclass C
	in
	    case super of
		NONE => ()
	      | SOME S =>
		if S = "java.lang.Object" then ()
		else if member C_ACCfinal (getCFlags S)
		then classErr (C ^ " extends final class " ^ S)
		else if member C_ACCinterface (getCFlags C)
		then if isInterface S then checkSuper S (C::done)
		     else classErr ("Interface " ^ C ^ " extends class " ^ S)
		else if isInterface S
		then classErr ("Class " ^ C ^ " extends interface " ^ S)
		else checkSuper S (C::done)
	end



(* The next function only checks the superficial structure of a class definition.
   Methods and fields must be checked immendiately prior to compilation since
   (for methods at least) typechecking generates type information which is required
   for compilation, and this information is discarded when the next method is
   typechecked. *)

fun typeCheckClassDef (cdef as CDEF(flags, C, super, interfaces, fdefs, mdefs, layout)) =
   let
       val () = Metadata.resetTypes ()  (* map from type names to integers for metadata *)

       val () = setCurrentClass C flags
       val () = checkCflags flags

       val () = if !strict then
		    (Classloader.saveUserClassInfo [cdef];
		     checkSuper C [];
		     if member C_ACCabstract flags then () (* Recall interfaces must be abstract*)
		     else checkInterfaces C)    
		else ()

         (* If we were compiling several classfiles at once we'd like to save the
            info for all of them (in Compile), but now we only parse them one at a time. *)

       fun fUniq l =
	   case l of [] => ()
		   | FDEF(_,ty,name,_)::t =>
		     if List.exists (fn FDEF(_,ty',name',_) => name' = name andalso ty' = ty)  t
		     then classErr ("Field " ^ name
				    ^ " is defined more than once")
		     else fUniq t

       val () = fUniq fdefs

       fun sameMethod name rty tys flags (MDEF(flags',rty',name',args,_)) =
	   name' = name 
	   andalso (map #1 args = tys)
	   andalso rty' = rty  
(*	   andalso member M_ACCstatic flags' = member M_ACCstatic flags*)

           (* FIX:  Java regards two methods as being indistinguishable if they
              have the same name and argument types.  JVM can distiguish two
              such methods if they have different return types or if
	      one is static and the other non-static. *)

       fun mUniq l =
	   case l of [] => ()
		   | MDEF(flags,rty,name,args,_)::t =>
		     if List.exists (sameMethod name rty (map #1 args) flags) t then
				 classErr ("Method " ^ name
					   ^ "(" ^ listToString tyToString "," (map #1 args) ^ ")"
					   ^ " is defined more than once")
			     else mUniq t

       val () = mUniq mdefs

   in
       ()
   end





(*  Here's the correct algorithm for method resolution.  I don't think we
    need this since it always looks both in superclasses and superinterfaces,
    but we usually know in advance whether we want a class or an interface. *)

(*
    fun resolveMethodI m rty tys intfs =
	let
	    fun doInterface J l done =
		if member J done then doList l done
		else
		    if not (isInterface J)
		    then err ("interface expected, but class " ^ J ^ "found")
		    else
			case seekMethod m rty tys J of
			    SOME m' => (vprint ("Found " ^ m ^ " in interface " ^ J ^ "\n");  (J,m'))
			  | NONE =>
			    case getSuperclass J of
				SOME J' => doInterface J' l (J::done)
			      | NONE => doList l (J::done)

	    and doList l done =
		case l of
		    [] => err ("Unable to resolve method " ^ m)
		  | I::t => doInterface I t done
	in
	    doList intfs []
	end

    fun resolveMethod m rty tys B =
	let
	    val () = resolveClass B

	    fun f C =
		let
		    val () = if member C_ACCinterface (getCFlags C)
			     then err ("class required,  but interface " ^ C ^  " found")
			     else ()
		in
		    case seekMethod m rty tys C of
			SOME m' => (vprint ("Found " ^ m ^ " in class " ^ C ^ "\n"); (C,m'))
		      | NONE =>
			case getSuperclass C of
			    SOME C' => f C'
			  | NONE => resolveMethodI m rty tys (getInterfaces B)
		end

		val (C,m') = f B

		val () = if member M_ACCabstract (methFlags m')
				 andalso (not (abstractClass C))
			 then err (m ^ " resolves to abstract method")
			 else checkMethodAccessible (!currentClass, C, m')
	in
	    C
	end

*)


(*

fun assingable S T =
    case S of 
	REFty s => 
	if isClass s then subclass S T
	else T = Object orelse subclass S T
      | ARRRAYty SC =>
	(
	 case T of
	     REFty t => 
	     if isClass t then t = "java.lang.Object"
	     else t = "java.lang.Cloneable" orelse t = "java.io.Serializable"
	   | ARRARty TC => assignable SC TC
	)
      | _ => FIX: narrowing

	    

    * If S is a class type:

          o If T is a class type, then S must be the same class as T, or S must be a subclass of T.

          o If T is an interface type, then S must implement interface T.

    * If S is an interface type:

          o If T is a class type, then T must be Object.

          o If T is an interface type, then T must be the same interface as S, or T must be a superinterface of S.

    * If S is an array type SC[], that is, an array of components of type SC:

          o If T is a class type, then T must be Object.

          o If T is an interface type, then T must be either Cloneable or java.io.Serializable.

          o If T is an array type TC[], that is, an array of components of type TC, then either

                + TC and SC must be the same primitive type, or

                + TC and SC are both reference types and type SC is assignable to TC. 

*)