(* Wrapper for standalone version of gdf *)

exception toyError of string;
exception gdfError of string;

fun usage() = TextIO.print "Usage: gdf [-c] [-t [syntax]] [-d outputdir] file1 [file2 ...]\n\
\   -d dir       use dir as output directory\n\
\   -c           print class name\n\
\   -t [syntax]  generate Isabelle .thy file as well [in given syntax]\n"

fun extend extn filename = (* Add extension if necessary *) (* Stolen from Camelot/src/Util.sml *)
   let
       val e = Path.ext filename;
   in
       if e = SOME extn
       then
	   filename
       else
	   Path.joinBaseExt {base=filename, ext=SOME extn}
   end



(* val doargs: string list -> string -> bool -> int option -> unit *)
(* thySyntax encoding = (printCertificate?, dataLayout, tag_offset, logic *)

fun doargs args dir printClassName (thySyntax as (p,d,tag,thy)) =
    case args of
	"-h"::t => usage ()

      | "-c"::t => doargs t dir true thySyntax

      | "-cp"::t =>
	let in
	    case t of
		[] => GrailUtils.exit "No classpath specified"
	      | h::u =>
		let
		    val () = Classloader.setClassPath h
		in
		    doargs u dir printClassName thySyntax
		end
	end

      | "-C"::t => doargs t dir printClassName (true,d,tag,thy)
			  (* enable certificate generation by setting 100 field in thySyntax *)

      | "-d"::t =>
	let in
	    case t of
		nil => raise (gdfError "no directory specified for classfile output")
	      | d::args' =>
	 	if not (Nonstdio.file_exists d) then raise (gdfError ("can't find directory " ^ d))
		else
		    if not (FileSys.isDir d) then raise (gdfError (d ^ " is not a directory"))
		    else let
        	            val dir = if (String.sub (d, (String.size d)-1)) = #"/"
	 		              then d else d ^ "/"
			in
			    doargs args' dir printClassName thySyntax
			end
	end

      | "-v"::t => (GrailUtils.verbose := true; doargs t dir printClassName thySyntax)

      | "-l"::t => doargs t dir printClassName (p,d+10,tag,thy)

      | "-L"::t  =>  (* set data layout *)
	let in
	    case t of
		nil => usage()
	      | layoutStr::t' =>
		let
		    val thySyntax = (p,d,tag,thy)
		    val l = case (Int.fromString layoutStr) of
				NONE => 0
			      | (SOME x) => x
		in
		    doargs t' dir printClassName (p,l,tag,thy) (* set layout *)
		end
	end

      | "-n"::t => (Typecheck.setStrictTypeChecking false; doargs t dir printClassName thySyntax)

      | "-s"::t => (ToyGrailAbsyn.shut_up := true ; doargs t dir printClassName thySyntax)

      | "-strict"::t => (Typecheck.setStrictTypeChecking true; doargs t dir printClassName thySyntax)

      | "-t"::t =>
	let
	in case t of
	       nil => usage()
	     | z::args' =>
	       if (String.isPrefix "-" z) then doargs (z::args') dir printClassName (p,d,tag,5)
						      (* 5 is default syntax *)
	       else
		   case z of
		       (* default:1 toy:2 bcl:3 *)
		       "toy" => doargs args' dir printClassName (p,d,tag, 2)
				       (* use ToyGrail syntax *)
		     | "bcl" => doargs args' dir printClassName (p,d,tag, 3)
				       (* use BytecodeLogic syntax *)
		     | "dal" => doargs args' dir printClassName (p,d,tag, 4)
				       (* use DA Logic syntax *)
		     | "vcg" => doargs args' dir printClassName (p,d,tag, 5)
				       (* use VCG Logic syntax *)
		     | _ => doargs (z::args') dir printClassName (p,d,tag, 5)
				   (*| _ => raise (gdfError (d ^ " is not a recognised logic format"))*)
	end

      | "-T"::t => (* set tag offset *)
	 let in
	     case t of
		 nil => usage()
	       | tagStr::t' =>
		 let
		     val tag' = case (Int.fromString tagStr) of
				    NONE => 0
				  | (SOME x) => x
		 in
		     doargs t' dir printClassName (p,d,tag',thy) (* set layout *)
		 end
	 end

      | "-test"::t => (Compile.testing := true; doargs t dir printClassName thySyntax)

      | "-unsuper"::t => (Compile.setACCsuper := false; doargs t dir printClassName thySyntax)

      | "-w"::t =>
	let (* set verbose level *)
	in case t of
	       nil => usage()
	     | d::t' =>
	       if (String.isPrefix "-" d)
	       then (ToyGrailAbsyn.verbose_level := ToyGrailAbsyn.VERB_MAX ;
		     doargs t dir printClassName thySyntax)
	       else case Int.fromString d of
			NONE => (ToyGrailAbsyn.verbose_level := ToyGrailAbsyn.VERB_MAX ;
				 doargs t dir printClassName thySyntax)
		      | (SOME v) => (ToyGrailAbsyn.verbose_level := v ;
				     doargs t' dir printClassName thySyntax)
	end

      | [] => (* No file name *)
	      let
		  val cdef = MainParser.parseStdIn ()
		  val cname = Compile.compile "std_in" cdef dir thySyntax
					      (Binarymap.mkDict (String.compare))
					      (* empty Dictionary -> no rich type predicates *)
	      in
		  if printClassName then
		      TextIO.print (cname ^ "\n")
		  else (* print a (not very) helpful message *)
			   TextIO.print ("Compiled " ^ "std_in" ^ "\n")
	      end


      |  _ =>
	let
	    fun compileFile filename =
		let
		    val cdef = MainParser.parse filename
		    val cname = Compile.compile filename cdef dir thySyntax
				(Binarymap.mkDict (String.compare))
				(* empty Dictionary -> no rich type predicates *)
		in
		    if printClassName then
			TextIO.print (cname ^ "\n")
		    else (* print a (not very) helpful message *)
			     TextIO.print ("Compiled " ^ filename ^ "\n")
		end
	in
	    app compileFile (map (extend "gr") args)
	end

val _ = doargs (CommandLine.arguments()) "" false  (false,0,0,0)
         (* default: don't print cert, data layout 0, tag offset 0, thySynt 0 *)
	   handle gdfError s => TextIO.print ("Error: " ^ s ^"\n")

