(* Int64.sml
 *
 * Peter Bertelsen
 * November 1997
 *)

datatype int = INT of Word64.word

exception Int64Overflow of string

local
    val min = Word8Vector.fromList [0wx80, 0w0, 0w0, 0w0, 0w0, 0w0, 0w0, 0w0]
    val max = Word8Vector.fromList [0wx7f, 0wxff, 0wxff, 0wxff,
				    0wxff, 0wxff, 0wxff, 0wxff]
in
    val minInt = INT (valOf(Word64.fromBytes min))
    val maxInt = INT (valOf(Word64.fromBytes max))
end

fun fromInt k = INT (Word64.fromInt k)

fun toInt (INT n) = Word64.toIntX n
    handle Word64.Word64Overflow s => raise Int64Overflow s

fun toString w = (Int.toString (toInt w))
    handle Int64Overflow s => s

fun fromBytes bs =
    (case Word64.fromBytes bs of
	 NONE   => NONE
       | SOME n => SOME (INT n))

fun toBytes (INT n) = Word64.toBytes n

fun compare (INT x, INT y) = Word64.compareX(x, y)

fun emit out (INT n) = Word64.emit out n

fun scan src =
    (case Word64.scan src of
	 NONE   => NONE
       | SOME n => SOME (INT n))



(* Arithmetic *)

local

open Word8Vector
fun sign b = Word8.andb(0wx80, sub(b,0))

fun add (u,v) =
    let
	val a = toBytes u
	val b = toBytes v
	fun h n prev acc =
	    if n < 0 then (fromList (List.map Word8.fromInt acc))
	    else let val p = Word8.toInt (sub (a,n))
		     val q = Word8.toInt (sub (b,n))
		     val i = p+q + prev div 256
		 in
		     h (n-1) i (i::acc)
		 end
	val r = h 7 0 []
	val () = let val s = sign a
		 in if sign b <> s then () (* No overflow *)
		    else if sign r = s then ()
		    else raise Overflow
		 end
    in
	valOf (fromBytes r)
    end


fun subtract (u,v) =
    let
	val a = toBytes u
	val b = toBytes v
	fun h n prev acc =
	    if n < 0 then (fromList (List.map Word8.fromInt acc))
	    else let val p = Word8.toInt (sub (a,n))
		     val q = Word8.toInt (Word8.notb (sub (b,n)))
		     val i = p+q + prev div 256
		 in
		     h (n-1) i (i::acc)
		 end
	val r = h 7 256 []
	val () = let val s = sign a
		 in if sign b = s then () (* No overflow *)
		    else if sign r = s then ()
		    else raise Overflow
		 end
    in
	valOf (fromBytes r)
    end

in

fun negate u =
    let
	val bs = toBytes u
	val u' = valOf(fromBytes (map Word8.notb (toBytes u)))
    in
	add(u', fromInt 1) (* ie, (not u)+1 *)
    end

val vecToList = foldr op:: []

fun toString n =
    let
	val b = toBytes n
	val b0 = sub (b,0)
    in
	if Word8.< (b0, 0wx80) then IntInf.toString (1, vecToList b)
	else if n = minInt then IntInf.toString (~1, vecToList b)
	else IntInf.toString (~1, vecToList (toBytes (negate n)))
    end

fun u + v = add(u,v)      handle Overflow => raise Int64Overflow (toString u ^ " + " ^ toString v)
fun u - v = subtract(u,v) handle Overflow => raise Int64Overflow (toString u ^ " - " ^ toString v)

end

fun fromString s =
    let
	val (sign, b) = IntInf.fromString 8 s
	val b0 = hd b
    in
	if b = [0wx80, 0w0, 0w0, 0w0, 0w0, 0w0, 0w0, 0w0] andalso sign < 0 then minInt
	else if Word8.>= (b0, 0wx80)
	then raise Int64Overflow ("Integer constant out of range: " ^ s)
	else
	    let
		val b' = valOf (fromBytes (Word8Vector.fromList b))
	    in
		if sign < 0 then negate b'
		else b'
	    end
    end

val fromHexString  = INT o Word64.fromHexString
fun toHexString (INT w) = Word64.toHexString w

