(* IntInf.sml *)

(* Kenneth MacKenzie, November 2005:
   Base conversions for arbitrary precision integers.
   A base-n integer is represented by a list of integers, one for each digit.
   The main reason for this is to allow us to input 32- and 64-bit literal
   integers for conversion to Int32.int and Int64.int *)

exception IntInfError of string

fun convert sourceBase targetBase n =
let
    fun divmod' l m acc =
	case l of
	    [] => (rev acc, m)
	  | h::t =>
	    let
 		val n = sourceBase*m + h
	    in  (* don't ignore inner zeros *)
		divmod' t (n mod targetBase) (n div targetBase :: acc)
	    end

    fun divmod l m acc =
	case l of
	    [] => (rev acc, m)
	  | h::t =>
	    let
 		val n = sourceBase*m + h
	    in
		if n < targetBase then divmod t n acc  (* ignore leading zeros *)
		else divmod' t (n mod targetBase) (n div targetBase :: acc)
	    end

    fun convert' n acc =
	case n of
	    [] => acc
	  | _ =>
	    let
		val (q,r) = divmod n 0 []
	    in
		convert' q (r::acc)
	    end
in
    convert' n []
end


(* Conversions from decimal string to byte list.  Internally, decimal
   numbers are represented by lists of integers in the range 0..9 *)


fun pad l x n =
    if n <= 0 then l
    else pad (x::l) x (n-1)

val zero = ord #"0"

fun toDecDigits l =
    if List.all Char.isDigit l
    then map (fn x => ord x - zero) l
    else raise IntInfError "Bad decimal"

fun fromString n s =
    let
	val l = explode s
	val (sign, b) =
	    case l of
		[] => (1,[])
	      | #"-"::t =>
		let val digits = toDecDigits t
		in if List.all (fn x => x=0) digits then (1, [0])
		   else (~1, convert 10 256 digits)
		end
	      | #"+"::t => (1, convert 10 256 (toDecDigits t))
	      | _ => (1,convert 10 256 (toDecDigits l))
	val b' = map Word8.fromInt b
	val r = length b'
    in
	if r > n then raise IntInfError "Integer constant too big"
	else (sign, pad b' 0wx0 (n-r))
    end


fun fromDecDigits l =
	implode (map (fn x => chr (x+zero)) l)

fun toString (sign, b) =
    let
	val b' = map Word8.toInt b
	val digits = fromDecDigits (convert 256 10 b')
    in
	if sign >= 0 (* orelse List.all (fn x => x=0) b' *)
	then digits
	else "-" ^ digits
    end

